""" Class for interaction with sqlite """
import base64
import copy
import io
import json
import logging
import os
import re
import shutil
import sqlite3
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Optional, Union
import pandas as pd
from anytree import Node
from PIL import Image
from .fixedStringsJson import SQLiteTranslation, defaultDefinitions, defaultDocTypes, defaultSchema
from .miscTools import hierarchy
from .textTools.stringChanges import camelCase, createDirName, outputString, tracebackString

MAIN_ORDER     =['id'  ,'name','user','type','dateCreated','dateModified','gui',      'client','shasum','image','content','comment','externalId','dateSync']
MAIN_TYPE      =['TEXT','TEXT','TEXT','TEXT','TEXT',       'TEXT',        'varchar(2)','TEXT',  'TEXT',  'TEXT', 'TEXT',   'TEXT',   'TEXT',     'TEXT']
DOC_TYPES      =['docType', 'PURL','title','icon','shortcut','view']
DOC_TYPE_SCHEMA=['docType', 'class', 'idx', 'name', 'unit', 'mandatory', 'list']

class SqlLiteDB:
  """
  Class for interaction with sqlite
  """
  def __init__(self, resetDataHierarchy:bool=False, basePath:Path=Path()):
    """
    Args:
      configuration (dict): configuration of GUI elements
      resetDataHierarchy (bool): reset dataHierarchy
      basePath (Path): path of project group
    """
    self.connection = sqlite3.connect(basePath/'pastaELN.db', check_same_thread=False)
    self.basePath   = basePath
    self.cursor     = self.connection.cursor()
    self.dataHierarchyInit(resetDataHierarchy)
    # main table
    self.createSQLTable('main',            MAIN_ORDER,                                  'id', MAIN_TYPE)
    # branches table
    self.createSQLTable('branches', ['id','idx','stack','child','path','show'],'id, idx', ['TEXT','INTEGER']*2+['TEXT']*2)
    # metadata table
    # - contains metadata of specific item
    # - key is generally not shown to user
    # - flattened key
    # - can be adopted by adv. user
    self.createSQLTable('properties',      ['id','key','value','unit'],                'id, key')
    # tables: each item can have multiple of these: tags, qrCodes
    self.createSQLTable('tags',            ['id','tag'],                               'id, tag')
    self.createSQLTable('qrCodes',         ['id','qrCode'],                            'id, qrCode')
    # list of changes to attachments
    self.createSQLTable('attachments',     ['id','location','date','guest','remark','user'], 'id, location, date')
    # list of changes to all documents: gives history
    self.createSQLTable('changes',         ['id','date','change'],                     'id, date')
    return


  def dataHierarchyInit(self, resetDataHierarchy:bool=False) -> None:
    """
    create data hierarchy in sqlite database

    Args:
      resetDataHierarchy (bool): recreate the data hierarchy from default state
    """
    self.cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
    tables = [i[0] for i in self.cursor.fetchall()] # all tables
    self.connection.commit()
    # check if default documents exist and create
    if 'docTypes' not in tables or resetDataHierarchy:
      if 'docTypes' in tables:
        print('Info: remove old docTypes')
        self.cursor.execute('DROP TABLE docTypes')
      self.createSQLTable('docTypes',  DOC_TYPES,    'docType')
      command = f"INSERT INTO docTypes ({', '.join(DOC_TYPES)}) VALUES ({', '.join(['?']*len(DOC_TYPES))});"
      self.cursor.executemany(command, defaultDocTypes)
      # docTypeSchema table (see below)
      # - define the key of the metadata table, units, icons, ...
      self.createSQLTable('docTypeSchema', DOC_TYPE_SCHEMA,  'docType, class, idx')
      command = f"INSERT INTO docTypeSchema ({', '.join(DOC_TYPE_SCHEMA)}) VALUES ({', '.join(['?']*len(DOC_TYPE_SCHEMA))});"
      self.cursor.executemany(command, defaultSchema)
      # definitions of all the keys (those from the docTypeSchema and those of the properties table)
      self.createSQLTable('definitions',     ['key','long','PURL'],                       'key')
      command =  'INSERT INTO definitions VALUES (?, ?, ?);'
      self.cursor.executemany(command, defaultDefinitions)
      self.connection.commit()
    return

  def dataHierarchyChangeView(self, docType:str, columns:list[str]) -> None:
    """ Set different view of docType in data hierarchy

    Args:
      docType (str): docType
      columns (list): list of columns
    """
    self.cursor.execute(f"UPDATE docTypes SET view='{','.join(columns)}' WHERE docType = '{docType}'")
    self.connection.commit()
    return


  def createSQLTable(self, name:str, columns:list[str], primary:str, colTypes:Optional[list[Any]]=None) -> list[str]:
    """
    Create a table in the sqlite system

    Args:
      name (str): name of table
      columns (list): list of column names
      primary (str): primary keys as comma-separated
      colTypes (list): optional list of data-types of the columns
    """
    if colTypes is None:
      colTypes = ['TEXT']*len(columns)
    colText = ', '.join(f'{i} {j}' for i,j in zip(columns, colTypes))
    self.cursor.execute(f"CREATE TABLE IF NOT EXISTS {name} ({colText}, PRIMARY KEY ({primary}))")
    self.cursor.execute(f"SELECT * FROM {name}")
    columnNames = list(map(lambda x: x[0], self.cursor.description))
    if columnNames != columns:
      logging.error('Difference in sqlite table: %s', name)
      logging.error(columnNames)
      logging.error(columns)
      raise ValueError(f'SQLite table difference: {name}')
    return columnNames


  def dataHierarchy(self, docType:str, column:str, group:str='') -> list[Any]:
    """
    get data hierarchy information, old name 'ontology', from database

    Args:
      docType (str): document typ
      column (str): column name as in meta, title, shortcut, ...
      group (str): group of metadata rows; if group not given: return all
    Returns:
      list: information inquired
    """
    ### return column information
    if not docType: #if all docTypes
      column = f', {column}' if column else ''
      self.cursor.execute(f"SELECT docType {column} FROM docTypes")
      results = self.cursor.fetchall()
      return [i[0] for i in results] if column=='' else results
    ### if metadata = docTypeSchema of data
    if column == 'meta':
      self.connection.row_factory = sqlite3.Row  #default None
      cursor = self.connection.cursor()
      if group:
        cursor.execute(f"SELECT * FROM docTypeSchema WHERE docType == '{docType}' and class == '{group}'")
      else:
        cursor.execute(f"SELECT * FROM docTypeSchema WHERE docType == '{docType}'")
      return [dict(i) for i in cursor.fetchall()]
    if column == 'metaColumns':
      self.cursor.execute(f"SELECT DISTINCT class FROM docTypeSchema WHERE docType == '{docType}'")
      return [i[0] for i in self.cursor.fetchall()]
    # if specific docType
    self.cursor.execute(f"SELECT {column} FROM docTypes WHERE docType == '{docType}'")
    result = self.cursor.fetchone()
    if result is None:
      return []
    if column=='meta':
      return json.loads(result[0])
    elif column=='view':
      return result[0].split(',')
    return result


  def exit(self) -> None:
    """
    Shutting down things
    """
    self.connection.close()
    return


  def getDoc(self, docID:str, noError:bool=False) -> dict[str,Any]:
    """
    Wrapper for get from database function

    Args:
        docID (dict): document id
        noError (bool): False=report errors as they occur; True=do not report on errors

    Returns:
        dict: json representation of document
    """
    self.connection.row_factory = sqlite3.Row  #default None
    cursor = self.connection.cursor()
    cursor.execute(f"SELECT * FROM main WHERE id == '{docID}'")
    res = cursor.fetchone()
    if res is None:
      if not noError:
        print(f'**ERROR sqlite: could not get docID: {docID} | {tracebackString(True, docID)}')
      return {}
    doc = dict(res)
    self.cursor.execute(f"SELECT tag FROM tags WHERE id == '{docID}'")
    doc['tags'] = [i[0] for i in self.cursor.fetchall()]
    self.cursor.execute(f"SELECT qrCode FROM qrCodes WHERE id == '{docID}'")
    doc['qrCodes'] = [i[0] for i in self.cursor.fetchall()]
    for key in ['image', 'content','shasum','client','qrCodes']:
      if len(doc[key])==0:
        del doc[key]
    doc['type']= doc['type'].split('/')
    doc['gui'] = [i=='T' for i in doc['gui']]
    doc['branch'] = []
    # data ends with 'id' 'stack', 'child', 'path', 'show', 'dateModified'
    self.cursor.execute(f"SELECT * FROM branches WHERE id == '{docID}'")
    for dataI in self.cursor.fetchall():
      doc['branch'].append({'stack': dataI[2].split('/')[:-1],
                            'child': dataI[3],
                            'path':  None if dataI[4] == '*' else dataI[4],
                            'show':   [i=='T' for i in dataI[5]]})
    cmd = 'SELECT properties.key, properties.value, properties.unit, definitions.long, definitions.PURL, ' \
          'docTypeSchema.unit FROM properties LEFT JOIN definitions USING(key) '\
          "LEFT JOIN docTypeSchema ON properties.key = (docTypeSchema.class || '.' || docTypeSchema.name) "\
          f"WHERE properties.id == '{docID}'"
    self.cursor.execute(cmd)
    res = self.cursor.fetchall()
    metadataFlat:dict[str, tuple[str,str,str,str]]  = {i[0]:(    i[1],            #value
                ('' if i[2] is None else i[2]) or ('' if i[5] is None else i[5]), #unit (priority: property.unit)
                ('' if i[3] is None else i[3]),                                   #long
                ('' if i[4] is None else i[4])                                    #PURL
                ) for i in res}
    doc |= hierarchy(metadataFlat)
    return doc


  def saveDoc(self, doc:dict[str,Any]) -> dict[str,Any]:
    """
    Save to database
    - not helpful to convert _id to int since sqlite does not digest such long integer
      doc['id']  = int(doc['id'][2:], 16)
    - Example{
        'name': 'G200X',
        'user': 'somebody',
        'type': ['instrument'],
        'branch': {'stack': ['x-9f74f79c96754d4c9065435ddd466c56', 'x-04b5e323d3ae4364bcd310a3e5fd653e'],
                  'child': 9999, 'path': None, 'show': [True, True, True], 'op': 'c'},
        'id': 'i-180212be6c7d4365ac647f266c5698f1',
        'externalId': '',
        'dateCreated': '2024-07-31T22:03:39.530666',
        'dateModified': '2024-07-31T22:03:39.530727',
        'gui': [True, True],
        'comment': '',
        '.vendor': 'KLA',
        '.model': 'KLA G200X',
        'metaVendor': {'j': 258},
        'tags': []}

    Discussion on -branch['path']:
    - full path (from basePath) allows to easily create a view of all paths and search through them
      during each scan, which happens rather often
    - just the incremental path (file-name, folder-name) allows to easily change that if the user wants
      and not change all the children paths, too. However, the renaming of the folder is likely occurring
      less often.

    Args:
        doc (dict): document to save

    Returns:
        dict: json representation of submitted document
    """
    # print('\nsave\n'+'\n'.join([f'{k}: {v}' for k,v in doc.items()]))
    docOrg = copy.deepcopy(doc)
    # save into branch table
    self.cursor.execute(f"INSERT INTO branches VALUES ({', '.join(['?']*6)})",
                        [doc['id'],
                         0,
                         '/'.join(doc['branch']['stack']+[doc['id']]),
                         str(doc['branch']['child']),
                         '*' if doc['branch']['path'] is None else doc['branch']['path'],
                         ''.join(['T' if j else 'F' for j in doc['branch']['show']])])
    del doc['branch']
    # save into tags table
    self.cursor.executemany('INSERT INTO tags VALUES (?, ?);', zip([doc['id']]*len(doc['tags']), doc['tags']))
    del doc['tags']
    if 'qrCodes' in doc:
      cmd = 'INSERT INTO qrCodes VALUES (?, ?);'
      self.cursor.executemany(cmd, zip([doc['id']]*len(doc['qrCodes']), doc['qrCodes']))
      del doc['qrCodes']
    if 'content' in doc and len(doc['content'])>200:
      doc['content'] = doc['content'][:200]
    doc['type'] = '/'.join(doc['type'])
    doc['gui']  = ''.join(['T' if i else 'F' for i in doc['gui']])
    doc['client'] = tracebackString(False, 'save:'+doc['id'])
    docList = [doc[x] if x in doc else doc.get(f'.{x}','') for x in MAIN_ORDER]
    self.cursor.execute(f"INSERT INTO main VALUES ({', '.join(['?']*len(docList))})", docList)
    doc = {k:v for k,v in doc.items() if (k not in MAIN_ORDER and k[1:] not in MAIN_ORDER) or k == 'id'}

    def insertMetadata(data:dict[str,Any], parentKeys:str) -> None:
      parentKeys = f'{parentKeys}.' if parentKeys else ''
      cmd    = 'INSERT OR REPLACE INTO properties VALUES (?, ?, ?, ?);'
      cmdDef = 'INSERT OR REPLACE INTO definitions VALUES (?, ?, ?);'
      for key,value in data.items():
        key = str(key) if isinstance(key, int) else key
        if isinstance(value, dict):
          insertMetadata(value, f'{parentKeys}{key}')
        elif key.endswith(']') and ('[') in key:
          unit   = re.findall(r'\[\S+\]', key)[-1][1:-1]
          label  = key[:-len(unit)-2].strip()
          key    = camelCase(label)
          key    = key[0].lower()+key[1:]
          self.cursor.execute(cmd, [doc['id'], parentKeys+key, str(value), unit])
          self.cursor.execute(cmdDef, [parentKeys+key, label, ''])
        elif isinstance(value, list) and isinstance(value[0], dict) and value[0].keys() >= {'key', 'value', 'unit'}:
          self.cursor.executemany(cmd, zip([doc['id']]*len(value),      [parentKeys+key+'.'+i['key'] for i in value],
                                            [i['value'] for i in value], [i['unit'] for i in value]  ))
          self.cursor.executemany(cmdDef, zip([parentKeys+key+'.'+i['key'] for i in value],
                                            [i['label'] for i in value], [i['PURL'] for i in value]  ))
        elif isinstance(value, tuple) and len(value)==4:
          self.cursor.execute(cmd,    [doc['id'], parentKeys+key, value[0], value[1]])
          self.cursor.execute(cmdDef, [parentKeys+key, value[2], value[3]])
        elif str(value)!='':
          try:
            self.cursor.execute(cmd, [doc['id'], parentKeys+key, str(value), ''])
          except Exception:
            print('**ERROR ', cmd, [doc['id'], parentKeys+key, str(value), ''])
      self.connection.commit()
      return
    # properties
    metaDoc = {k:v for k,v in doc.items() if k not in MAIN_ORDER}
    insertMetadata(metaDoc, '')
    # save changes
    self.connection.commit()
    branch = copy.deepcopy(docOrg['branch'])
    del branch['op']
    docOrg['branch'] = [branch]
    return docOrg


  def updateDoc(self, dataNew:dict[str,Any], docID:str) -> dict[str,Any]:
    """
    Update document by
    - saving changes to oldDoc (revision document)
    - updating new-document concurrently
    - create a docID for oldDoc

    Key:Value
    - if value is None: do not change it;
    - if key does not exist: change it to empty, aka remove it

    Args:
        dataNew (dict): item to update
        docID (string):  id of document to change

    Returns:
        dict: json representation of updated document
    """
    if set(dataNew.keys()) == {'type','image'}: #if only type and image in update = change of extractor
      self.cursor.execute(f"UPDATE main SET type='{'/'.join(dataNew['type'])}', image='{dataNew['image']}' WHERE id = '{docID}'")
      self.connection.commit()
      return {'id':docID}
    dataNew['client'] = tracebackString(False, f'updateDoc:{docID}')
    if 'edit' in dataNew:     #if delete
      dataNew = {'id':dataNew['id'], 'branch':dataNew['branch'], 'user':dataNew['user'], 'externalId':dataNew['externalId'], 'name':''}
    changesDict:dict[str,Any]         = {}
    self.connection.row_factory = sqlite3.Row  #default None
    cursor = self.connection.cursor()

    # tags and qrCodes
    tagsNew= set(dataNew.pop('tags'))
    self.cursor.execute(f"SELECT tag FROM tags WHERE id == '{docID}'")
    tagsOld= {i[0] for i in self.cursor.fetchall()}
    if tagsOld.difference(tagsNew):
      cmd = f"DELETE FROM tags WHERE id == '{docID}' and tag == ?"
      self.cursor.executemany(cmd, [(i,) for i in tagsOld.difference(tagsNew)])
      changesDict['tags'] = ','.join(tagsOld)
    if tagsNew.difference(tagsOld):
      change = tagsNew.difference(tagsOld)
      self.cursor.executemany('INSERT INTO tags VALUES (?, ?);', zip([docID]*len(change), change))
      changesDict['tags'] = ','.join(tagsOld)
    qrCodesNew= set(dataNew.pop('qrCodes', []))
    self.cursor.execute(f"SELECT qrCode FROM qrCodes WHERE id == '{docID}'")
    qrCodesOld= {i[0] for i in self.cursor.fetchall()}
    if qrCodesOld.difference(qrCodesNew):
      cmd = f"DELETE FROM qrCodes WHERE id == '{docID}' and qrCode == ?"
      self.cursor.executemany(cmd, [(i,) for i in qrCodesOld.difference(qrCodesNew)])
      changesDict['qrCodes'] = ','.join(qrCodesOld)
    if qrCodesNew.difference(qrCodesOld):
      change = qrCodesNew.difference(qrCodesOld)
      self.cursor.executemany('INSERT INTO qrCodes VALUES (?, ?);', zip([docID]*len(change), change))
      changesDict['qrCodes'] = ','.join(qrCodesOld)
    # separate into main and properties
    mainNew    = {key: dataNew.pop(key) for key in MAIN_ORDER if key in dataNew}
    branchNew  = dataNew.pop('branch',{})
    # read branches and identify changes
    cursor.execute(f"SELECT stack, child, path, show FROM branches WHERE id == '{docID}'")
    branchOld = [dict(i) for i in cursor.fetchall()]
    branchOld[0]['show'] = [i=='T' for i in branchOld[0]['show']]
    branchOld[0]['stack'] = branchOld[0]['stack'].split('/')[:-1]
    # print(f'do something with branch: \nnew {branchNew} \nold: {branchOld}')
    if branchNew:
      op      = branchNew.pop('op')             #operation
      oldPath = branchNew.pop('oldPath',None)
      if not branchNew['path']:
        branchNew['path'] = branchOld[0]['path']
      if branchNew not in branchOld:            #only do things if the newBranch is not already in oldBranch
        changesDict['branch'] = branchOld.copy()
        idx = None
        for branch in branchOld:                #check if the path is already exist, then it is only an update
          if op=='c' and branch['path']==branchNew['path']:
            op='u'
        if op=='c':    #create, append
          if isinstance(branchNew['stack'], str):
            raise ValueError('Should be list')
          branchNew['show'] = self.createShowFromStack(branchNew['stack'])
          self.cursor.execute(f"INSERT INTO branches VALUES ({', '.join(['?']*6)})",
                        [docID,
                         len(branchOld),
                         '/'.join(branchNew['stack']+[docID]),
                         branchNew['child'],
                         '*' if branchNew['path'] is None else branchNew['path'],
                         ''.join(['T' if j else 'F' for j in branchNew['show']])])
        elif op=='u':  #update
          if oldPath is not None:              # search by using old path
            for branch in branchOld:
              if branch['path'].startswith(oldPath):
                if   os.path.basename(branch['path']) == mainNew['name'] and \
                     os.path.basename(str(branchNew['path']))!='':
                  mainNew['name'] = os.path.basename(str(branchNew['path']))
                branch['path'] = branch['path'].replace(oldPath ,branchNew['path'])
                branch['stack']= branchNew['stack']
                branch['show'] = self.createShowFromStack(branchNew['stack'], branch['show'][-1])
                break
          else:
            idx = 0
            branchOld[idx] = branchNew           #change branch 0 aka the default
          if idx is None:                          # create new branch: should not happen here
            raise ValueError(f'sqlite.2: idx unset: {mainNew["id"]} {mainNew["name"]}')
          self.cursor.execute(f"UPDATE branches SET stack='{'/'.join(branchOld[idx]['stack']+[docID])}', "
                              f"path='{branchOld[idx]['path']}', child='{branchOld[idx]['child']}', "
                              f"show='{''.join(['T' if j else 'F' for j in branchOld[idx]['show']])}' "
                              f"WHERE id = '{docID}' and idx = {idx}")
        elif op=='d':  #delete
          branchOld = [branch for branch in branchOld if branch['path']!=branchNew['path']]
        else:
          raise ValueError(f'sqlite.1: unknown branch op: {mainNew["id"]} {mainNew["name"]}: {branchNew} of doc {dataNew}')

    # read properties and identify changes
    self.cursor.execute(f"SELECT key, value FROM properties WHERE id == '{docID}'")
    dataOld = {i[0]:i[1] for i in self.cursor.fetchall()}
    for key,value in dataNew.items():
      if isinstance(value, dict):
        for subKey, subValue in value.items():
          longKey = f'{key}.{subKey}'
          if isinstance(subValue, tuple):
            subValue = subValue[0]
          if longKey in dataOld and subValue!=dataOld[longKey]:
            self.cursor.execute(f"UPDATE properties SET value='{subValue}' WHERE id = '{docID}' and key = '{longKey}'")
            changesDict[longKey] = dataOld[longKey]
          elif longKey not in dataOld:
            cmd = 'INSERT INTO properties VALUES (?, ?, ?, ?);'
            self.cursor.execute(cmd ,[docID, longKey, subValue, ''])
          if longKey in dataOld:
            del dataOld[longKey]
      elif isinstance(value, str) and '.' in key:
        if key in dataOld and value!=dataOld[key]:
          self.cursor.execute(f"UPDATE properties SET value='{value}' WHERE id = '{docID}' and key = '{key}'")
          changesDict[key] = dataOld[key]
        elif key not in dataOld:
          cmd = 'INSERT INTO properties VALUES (?, ?, ?, ?);'
          self.cursor.execute(cmd ,[docID, key, value, ''])
      elif isinstance(value, tuple) and len(value)==4:
        if key in dataOld and value[0]!=dataOld[key]:
          self.cursor.execute(f"UPDATE properties SET value='{value[0]}' WHERE id = '{docID}' and key = '{key}'")
          changesDict[key] = dataOld[key]
        elif key not in dataOld:
          cmd = 'INSERT INTO properties VALUES (?, ?, ?, ?);'
          self.cursor.execute(cmd ,[docID, key, value[0], value[1]])
      else:
        logging.error('Property is not a dict, ERROR %s %s',key, value)
    if set(dataOld.keys()).difference(dataNew.keys()):
      cmd = f"DELETE FROM properties WHERE id == '{docID}' and key == ?"
      properties = [(i,) for i in set(dataOld.keys()).difference(dataNew.keys())]
      self.cursor.executemany(cmd, properties)
      changesDict |= dataOld
    # read main and identify if something changed
    cursor.execute(f"SELECT * FROM main WHERE id == '{docID}'")
    mainOld = dict(cursor.fetchone())
    mainOld['type']= mainOld['type'].split('/')
    changesDB: dict[str,dict[str,str]] = {'main': {}}
    for key in ('name','user','type','dateModified','dateSync','client','image','content','comment'):
      if key in mainNew and mainOld[key]!=mainNew[key]:
        changesDB['main'][key] = '/'.join(mainNew[key]) if key=='type' else mainNew[key].translate(SQLiteTranslation)
        changesDict[key] = mainOld[key]
    # save change content in database: main and changes are updated
    if set(changesDict.keys()).difference(('dateModified','client','user')):
      changeString = ', '.join([f"{k}='{v}'" for k,v in changesDB['main'].items()])
      self.cursor.execute(f"UPDATE main SET {changeString} WHERE id = '{docID}'")
      if 'name' not in changesDict or changesDict['name']!='new item':  #do not save initial change from new item
        self.cursor.execute('INSERT INTO changes VALUES (?,?,?)', [docID, datetime.now().isoformat(), json.dumps(changesDict)])
      self.connection.commit()
    return mainOld | mainNew | {'branch':branchOld, '__version__':'short'}


  def updateBranch(self, docID:str, branch:int, child:int, stack:list[str]=[],
                   path:Optional[str]='', **kwargs:str) -> tuple[Optional[str], Optional[str]]:
    """
    Update document by updating the branch

    Args:
      docID (string):  id of document to change
      branch (int):  index of branch to change. If -1, then append to end. If -2, then delete
      child (int):  new number of child
      stack (list):  new list of ids
      path (str): new path; None is acceptable
      kwargs (str): pathOld and stackOld can be used to identify branch to be updated

    Returns:
      str, str: old path, new path
    """
    #convert into db style
    path = '*' if path is None else path
    if branch == -2: #delete this path
      self.cursor.execute(f"DELETE FROM branches WHERE id == '{docID}' and path == '{path}'")
      self.connection.commit()
      # test if there is a branch remaining, if not delete document
      self.cursor.execute(f'SELECT id FROM branches WHERE id == "{docID}"')
      res = self.cursor.fetchall()
      if len(res)==0:
        self.remove(docID)
      return (path, None)

    if branch == -1: #append a new branch
      self.cursor.execute(f"SELECT idx FROM branches WHERE id == '{docID}'")
      idxOld = [i[0] for i in self.cursor.fetchall()]
      idxNew  = min(set(range(max(idxOld)+2)).difference(idxOld))
      show  = self.createShowFromStack(stack)
      self.cursor.execute(f"INSERT INTO branches VALUES ({', '.join(['?']*6)})",
                  [docID, idxNew, '/'.join(stack+[docID]), str(child), path, show])
      self.connection.commit()
      return (None, None if path=='*' else path)

    # modify existing branch
    if 'pathOld' in kwargs:
      pathOld = kwargs['pathOld']
      cmd = f"SELECT path, stack, show FROM branches WHERE path == '{pathOld}'"
    elif 'stackOld' in kwargs:
      stackOld = '/'.join(kwargs['stackOld'])
      cmd = f"SELECT path, stack, show FROM branches WHERE stack == '{stackOld}'"
    else:
      cmd = f"SELECT path, stack, show FROM branches WHERE id == '{docID}' and idx == {branch}"
    self.cursor.execute(cmd)
    reply = self.cursor.fetchone()
    if reply is None:
      raise ValueError(f"FAILED AT: {cmd}")
    pathOld, stackOld, showOld = reply
    stack = stack or stackOld.split('/')[:-1]           # stack without current id
    if pathOld=='*':
      path = '*'
    elif path=='':
      self.cursor.execute(f"SELECT type, name FROM main WHERE id == '{docID}'")
      docType, name = self.cursor.fetchall()[0]
      if docType.startswith('x'):
        name = createDirName(name,docType,child)
      path = ((Path(pathOld).parent)/name).as_posix()
    show  = self.createShowFromStack(stack, showOld[-1])
    cmd = f"UPDATE branches SET stack='{'/'.join(stack+[docID])}', child={child}, path='{path}', show='{show}' "\
          f"WHERE path = '{pathOld}' and stack = '{stackOld}'"
    self.cursor.execute(cmd)
    self.connection.commit()
    # move content: folder and data and write .json to disk
    if pathOld!='*' and ':/' not in pathOld and path!='*' and path is not None:
      if not (self.basePath/pathOld).exists() and (self.basePath/path).exists():
        logging.debug('sqlite:updateBranch: dont move since already good')
      else:
        shutil.move(self.basePath/pathOld, self.basePath/path)
    if docID[0]=='x' and path is not None:
      with open(self.basePath/path/'.id_pastaELN.json', 'w', encoding='utf-8') as fOut:
        fOut.write(json.dumps(self.getDoc(docID)))
      self.updateChildrenOfParentsChanges(pathOld, path,  stackOld,'/'.join(stack+[docID]))
    return (None if pathOld=='*' else pathOld, None if path=='*' else path)


  def updateChildrenOfParentsChanges(self, pathOld:str, pathNew:str, stackOld:str, stackNew:str) -> None:
    """
    update children's paths

    Args:
      pathOld (str): old path of parent
      pathNew (str): new path of parent
      stackOld (str): old stack of parent
      stackNew (str): new stack of parent
    """
    cmd = f"SELECT path, stack, show, id, idx FROM branches WHERE stack LIKE '{stackOld}/%'"
    self.cursor.execute(cmd)
    children = self.cursor.fetchall()
    updatedInfo = []
    for (pathIOld, stackIOld, showIOld, docID, idx) in children:
      pathINew  = pathNew+pathIOld[len(pathOld):] if pathIOld!='*' and pathIOld.startswith(pathOld) else pathIOld
      stackINew = stackNew+stackIOld[len(stackOld):]                           if stackNew else stackIOld
      showINew  = self.createShowFromStack(stackIOld.split('/'), showIOld[-1]) if stackNew else showIOld
      updatedInfo.append((pathINew, stackINew, showINew, docID, idx))
    cmd = 'UPDATE branches SET path=?, stack=?, show=? WHERE id == ? and idx == ?'
    self.cursor.executemany(cmd, updatedInfo)
    self.connection.commit()
    return


  def createShowFromStack(self, stack:list[str], currentShow:str='T') -> str:
    """
    For branches: create show entry in the branches by using the stack
    - should be 1 longer than stack
    - check parents if hidden, then this child is hidden too

    Args:
      stack (list): list of ancestor docIDs '/' separated
      currentShow (str): current show-indicator of this item

    Returns:
      str: string of show = list of bool
    """
    show = len(stack)*['T'] + [currentShow]
    for idx, docID in enumerate(stack):
      self.cursor.execute(f"SELECT show FROM branches WHERE id == '{docID}'")
      if self.cursor.fetchone()[0][-1] == 'F':
        show[idx] = 'F'
    return ''.join(show)

  def remove(self, docID:str, stack:str='') -> dict[str,Any]:
    """
    remove doc from database: temporary for development and testing

    Args:
      docID (string): id of document to remove
      stack (str): stack of branch to delete, if there are multiple branches

    Returns:
      dict: document that was removed
    """
    doc = self.getDoc(docID)
    if len(doc['branch'])>1 and stack:  #only remove one branch
      stack = stack[:-1] if stack.endswith('/') else stack
      self.cursor.execute(f"DELETE FROM branches WHERE id == '{docID}' and stack LIKE '{stack}%'")
    else:                               #remove everything
      doc.pop('image','')
      doc.pop('content','')
      self.cursor.execute(f"DELETE FROM main WHERE id == '{docID}'")
      self.cursor.execute(f"DELETE FROM branches WHERE id == '{docID}'")
      self.cursor.execute(f"DELETE FROM properties WHERE id == '{docID}'")
      self.cursor.execute(f"DELETE FROM tags WHERE id == '{docID}'")
      self.cursor.execute(f"DELETE FROM qrCodes WHERE id == '{docID}'")
      self.cursor.execute(f"DELETE FROM attachments WHERE id == '{docID}'")
      self.cursor.execute('INSERT INTO changes VALUES (?,?,?)', [docID, datetime.now().isoformat(), json.dumps(doc)])
    self.connection.commit()
    return doc


  def initAttachment(self, docID:str, name:str, docType:str) -> None:
    """ initialize an attachment by defining its name and type

    Args:
      docID (str): document / instrument on which attachments are installed
      name (str): description of attachment location
      docType (str): document type what can be attached. Use empty string if these are remarks, i.e. no items are attached
    """
    cmd = 'INSERT INTO attachments VALUES (?,?,?,?,?,?)'
    self.cursor.execute(cmd, [docID, name, '', docType, '', ''])
    self.connection.commit()
    return


  def addAttachment(self, docID:str, name:str, content:dict[str,Any]) -> None:
    """
    Update document by adding attachment (no new revision)

    Args:
        docID (string):  id of document to change
        name (string): attachment name to add to
        content (dict): dictionary of content to be added (should include user,date,docID,remark)

    """
    cmd = 'INSERT INTO attachments VALUES (?,?,?,?,?,?)'
    self.cursor.execute(cmd, [docID, name, content['date'], content['docID'], content['remark'], content['user']])
    self.connection.commit()
    return


  def getView(self, thePath:str, startKey:Optional[str]=None, preciseKey:Optional[str]=None) -> pd.DataFrame:
    """
    Wrapper for getting view function

    Args:
        thePath (string): path to view
        startKey (string): / separated; if given, use to filter output, everything that starts with this key
        preciseKey (string): if given, use to filter output. Match precisely

    Returns:
        df: data-frame with human-readable column names
    """
    if (startKey is not None and ' ' in startKey) or (preciseKey is not None and ' ' in preciseKey):
      raise ValueError(f'key cannot have a space inside |{startKey}|{preciseKey}|')  #TODO: remove this safeguard in 2026
    allFlag = False
    if thePath.endswith('All'):
      thePath = thePath.removesuffix('All')
      allFlag = True
    viewType, docType = thePath.split('/')
    if viewType=='viewDocType':
      viewColumns = self.dataHierarchy(docType, 'view')
      viewColumns = viewColumns+['id'] if viewColumns else ['name','tags','comment','id']
      textSelect = ', '.join([f'main.{i}' for i in viewColumns if i in MAIN_ORDER or i[1:] in MAIN_ORDER])
      cmd = f"SELECT {textSelect} FROM main INNER JOIN branches USING(id) WHERE main.type LIKE '{docType}%'"
      if not allFlag:
        cmd += r" and NOT branches.show LIKE '%F%'"
      if startKey:
        cmd += f" and branches.stack LIKE '{startKey}%'"
      df      = pd.read_sql_query(cmd, self.connection, index_col='id', ).fillna('')
      #clean main df
      df      = df[~df.index.duplicated(keep='first')]
      if 'image' in viewColumns:
        df['image'] = str(len(df['image'])>1)
      # add: tags, qrCodes, parameters
      if 'tags' in viewColumns:
        cmd = 'SELECT main.id, tags.tag FROM main INNER JOIN tags USING(id) INNER JOIN branches USING(id) '\
              f"WHERE main.type LIKE '{docType}%'"
        if not allFlag:
          cmd += r" and NOT branches.show LIKE '%F%'"
        if startKey:
          cmd += f" and branches.stack LIKE '{startKey}%'"
        dfTags = pd.read_sql_query(cmd, self.connection, index_col='id').fillna('')
        dfTags = dfTags.groupby(['id'])['tag'].apply(lambda x: ', '.join(set(x))).reset_index().set_index('id')
        df = df.join(dfTags)
      if 'qrCodes' in viewColumns:
        cmd = 'SELECT main.id, qrCodes.qrCode FROM main INNER JOIN qrCodes USING(id) INNER JOIN branches '\
              f"USING(id) WHERE main.type LIKE '{docType}%'"
        if not allFlag:
          cmd += r" and NOT branches.show LIKE '%F%'"
        if startKey:
          cmd += f" and branches.stack LIKE '{startKey}%'"
        dfQrCodes = pd.read_sql_query(cmd, self.connection, index_col='id').fillna('')
        dfQrCodes = dfQrCodes.groupby(['id'])['qrCode'].apply(', '.join).reset_index().set_index('id')
        df = df.join(dfQrCodes)
      if metadataKeys:= [f'properties.key == "{i}"' for i in viewColumns if i not in MAIN_ORDER+['tags']]:
        cmd = 'SELECT main.id, properties.key, properties.value FROM main INNER JOIN properties USING(id) '\
              f"INNER JOIN branches USING(id) WHERE main.type LIKE '{docType}%' AND ({' OR '.join(metadataKeys)})"
        if not allFlag:
          cmd += r" and NOT branches.show LIKE '%F%'"
        if startKey:
          cmd += f" and branches.stack LIKE '{startKey}%'"
        dfParams = pd.read_sql_query(cmd, self.connection, index_col='id').fillna('')
        dfParams = dfParams.pivot(columns='key', values='value').reset_index().set_index('id')  # Pivot the DataFrame
        dfParams.columns.name = None                                                            # Flatten the columns
        df = df.join(dfParams)
      # final sorting of columns
      columnOrder = ['tag' if i=='tags' else 'qrCode' if i=='qrCodes'
                     else i[1:] if i.startswith('.') and i[1:] in MAIN_ORDER else i for i in viewColumns]
      df = df.reset_index().reindex(columnOrder, axis=1)
      df = df.rename(columns={i:i[1:] for i in columnOrder if i.startswith('.') })
      df = df.astype('str').fillna('')
      return df
    elif thePath=='viewHierarchy/viewHierarchy':
      cmd = 'SELECT branches.id, branches.stack, branches.child, main.type, main.name, main.gui, branches.idx '\
            f"FROM branches INNER JOIN main USING(id) WHERE branches.stack LIKE '{startKey}%'"
      if not allFlag:
        cmd += r" and NOT branches.show LIKE '%F%'"
      cmd += ' ORDER BY branches.stack'
      self.cursor.execute(cmd)
      results = self.cursor.fetchall()
      # value: [child, doc['type'], doc['name'], doc['gui'], branches.idx]
      results = [{'id':i[0], 'key':i[1],
                  'value':[i[2], i[3].split('/'), i[4], [j=='T' for j in i[5]], i[6]]} for i in results]
    elif thePath=='viewHierarchy/viewPaths':
      cmd = 'SELECT branches.id, branches.path, branches.stack, main.type, branches.child, main.shasum, branches.idx '\
            'FROM branches INNER JOIN main USING(id)'
      # JOIN and get type
      if startKey is not None:
        if startKey.endswith('/'):
          startKey = startKey.removesuffix('/')
        cmd += f" WHERE branches.path LIKE '{startKey}%'"
      elif preciseKey is not None:
        cmd += f" WHERE branches.path LIKE '{preciseKey}'"
      if not allFlag:
        if 'WHERE' in cmd:
          cmd += r" AND NOT branches.show LIKE '%F%'"
        else:
          cmd += r" WHERE NOT branches.show LIKE '%F%'"
      self.cursor.execute(cmd)
      results = self.cursor.fetchall()
      # value: [branch.stack, doc['-type'], branch.child, doc.shasum, idx]
      results = [{'id':i[0], 'key':i[1],
                  'value':[i[2], i[3].split('/')]+list(i[4:])} for i in results if i[1] is not None]
    elif viewType=='viewIdentify' and docType=='viewTags':
      return pd.read_sql_query('SELECT tags.tag, main.name, main.type, tags.id FROM tags INNER JOIN main', self.connection).fillna('')
    elif viewType=='viewIdentify':
      if docType=='viewQR':
        if startKey is None:
          cmd = 'SELECT qrCodes.id, qrCodes.qrCode, main.name FROM qrCodes INNER JOIN main USING(id)'
          if not allFlag:
            cmd += r" INNER JOIN branches USING(id) WHERE NOT branches.show LIKE '%F%'"
        else:
          raise ValueError('Not implemented sqlite l 599')
      elif docType=='viewSHAsum':
        if startKey is None:
          cmd = 'SELECT main.id, main.shasum, main.name FROM main'
          if not allFlag:
            cmd += r" INNER JOIN branches USING(id) WHERE NOT branches.show LIKE '%F%'"
        else:
          cmd = f"SELECT main.id, main.shasum, main.name FROM main INNER JOIN branches USING(id) WHERE shasum='{startKey}'"
          if not allFlag:
            cmd += r" and NOT branches.show LIKE '%F%'"
      else:
        raise ValueError('Invalid docType')
      self.cursor.execute(cmd)
      results = self.cursor.fetchall()
      results = [{'id':i[0], 'key':i[1].replace('/',' '), 'value':i[2]} for i in results if i[1] is not None]
    else:
      print('continue here with view', thePath, startKey, preciseKey)
      raise ValueError('Not implemented')
    return results


  def getHierarchy(self, start:str, allItems:bool=False) -> tuple[Node,bool]:
    """
    get hierarchy tree for projects, ...

    Args:
      start (str): start of the hierarchy (most parent)
      allItems (bool):  true=show all items, false=only non-hidden

    Returns:
      Node: hierarchy in an anytree
      bool: error occurred
    """
    path = 'viewHierarchy/viewHierarchyAll' if allItems else 'viewHierarchy/viewHierarchy'
    view = self.getView(path,    startKey=start)
    # create tree of folders: these are the only ones that have children
    dataTree = None
    nonFolders = []
    id2Node = {}
    error = False
    for item in view:
      docType = item['value'][1]
      if docType[0][0] != 'x':
        nonFolders.append(item)
        continue
      _id     = item['id']
      childNum, docType, name, gui, _ = item['value']
      if dataTree is None:
        dataTree = Node(id=_id, docType=docType, name=name, gui=gui, childNum=childNum)
        id2Node[_id] = dataTree
      else:
        parent = item['key'].split('/')[-2]
        if parent in id2Node:
          parentNode = id2Node[parent]
        else:
          parentNode, error = (dataTree, True)
          print(f'**ERROR** there is an error in the hierarchy tree structure with parent {parent} missing')
        subNode = Node(id=_id, parent=parentNode, docType=docType, name=name, gui=gui, childNum=childNum)
        id2Node[_id] = subNode
    # add non-folders into tree
    # print(len(nonFolders),'length: crop if too long')
    for item in nonFolders:
      _id     = item['id']
      childNum, docType, name, gui, _ = item['value']
      parentId = item['key'].split('/')[-2]
      if parentId in id2Node:
        parentNode = id2Node[parentId]
      else:
        outputString('print','error',f'repair branch table as parentID {parentId} is missing')
        parentNode, error = (dataTree,True)
      Node(id=_id, parent=parentNode, docType=docType, name=name, gui=gui, childNum=childNum)
    # sort children
    for parentNode in id2Node.values():
      children = parentNode.children
      childNums= [f'{i.childNum}{i.id}{idx}' for idx,i in enumerate(children)]
      parentNode.children = [x for _, x in sorted(zip(childNums, children))]
    return dataTree, error


  def hideShow(self, docID:str) -> None:
    """
    Toggle hide/show indicator of branch

    Args:
      docID (str): document id
    """
    def adoptShow(item:tuple[str,str,str,str]) -> tuple[str,str,str]:
      """ Adopt show string, internal mapping function

      Args:
        item (tuple): id, idx, stack, show

      Returns:
        tuple: show, id, idx
      """
      localDocID, idx, stack, show = item
      j = stack.split('/').index(docID)
      showL = list(show)
      showL[j] = 'T' if showL[j]=='F' else 'F'
      return (''.join(showL), localDocID, idx)
    cmd = f"SELECT id, idx, stack, show FROM branches WHERE stack LIKE '%{docID}%'"
    self.cursor.execute(cmd)
    changed = map(adoptShow, self.cursor.fetchall())
    cmd = 'UPDATE branches SET show=? WHERE id = ? and idx= ?'
    self.cursor.executemany(cmd, changed)
    self.connection.commit()
    return


  def setGUI(self, docID:str, guiState:list[bool]) -> None:
    """
    Set the gui state
    - 0: true=show details; false=hide details
    - 1: true=show children; false=hide children; only makes sense for folders: doctype = x1

    Args:
      docID (str): docID
      guiState (list): list of bool that show if document is shown
    """
    guiList = ''.join(['T' if i else 'F' for i in guiState])
    cmd = f"UPDATE main SET gui='{guiList}' WHERE id = '{docID}'"
    self.cursor.execute(cmd)
    self.connection.commit()
    return


  def checkDB(self, outputStyle:str='text', minimal:bool=False,
              repair:Union[None,Callable[[str],bool]]=None) -> str:
    """
    Check database for consistencies by iterating through all documents
    - only reporting, no repair
    - custom changes are possible with normal scan
    - no interaction with hard disk

    Args:
        outputStyle (str): output using a given style: see outputString
        minimal (bool): true=only show warnings and errors; else=also show information
        repair (function): auto-repair after asking user

    Returns:
        str: output
    """
    reply = ''
    if outputStyle=='html':
      reply += '<div align="right">'
    if not minimal:
      reply+= outputString(outputStyle,'h2','LEGEND')
      reply+= outputString(outputStyle,'perfect','Green: perfect and as intended')
      reply+= outputString(outputStyle,'ok',
                               'Blue: ok, can happen: empty files for testing, strange path for measurements')
      reply+= outputString(outputStyle,'h2','List all database entries')
    if outputStyle=='html':
      reply += '</div>'
      reply+= outputString(outputStyle,'h2','List all database entries')
    lostAndFoundProjId   = ''
    lostAndFoundProjPath = Path()
    if repair is not None:
      dfProjects = self.getView('viewDocType/x0')
      idProjects = dfProjects[dfProjects['name']=='Lost and Found']['id'].values
      if len(idProjects)==0:
        print('**ERROR: manually create "LostAndFound" project to allow full repair')
        return '**ERROR: manually create "LostAndFound" project to allow full repair\n'
      else:
        lostAndFoundProjId   = idProjects[0]
        lostAndFoundProjPath = Path('LostAndFound')  #by definition
    # tests
    self.cursor.execute('SELECT main.id, main.name FROM main WHERE id NOT IN (SELECT id FROM branches)')
    if res:= self.cursor.fetchall():
      errorStr = outputString(outputStyle,'error', f'Items with no branch: {", ".join(str(i) for i in res)}')
      if repair is None:
        reply+= errorStr
      elif repair(errorStr):
        for docID, name in res:
          self.cursor.execute(f"INSERT INTO branches VALUES ({', '.join(['?']*6)})",
            [docID, 0, f'{lostAndFoundProjId}/{docID}', 9999,
            (lostAndFoundProjPath/createDirName(name, 'x0', 0)).as_posix() if docID.startswith('x-') else '*',
            'TT'])
        self.connection.commit()

    cmd = 'SELECT id, main.type, branches.stack, branches.path, branches.child, branches.show, main.name '\
          'FROM branches INNER JOIN main USING(id)'
    self.cursor.execute(cmd)
    res = self.cursor.fetchall()
    reply += f'Number of documents: {len(res)}\n'
    for row in res:
      try:
        docID, docType, stack, path, child, _, name = row[0], row[1], row[2], row[3], row[4], row[5], row[6]
      except ValueError:
        reply+= outputString(outputStyle,'error','dch03a: branch data has strange list')
        continue
      if docType.count('/')>5:
        reply+= outputString(outputStyle,'error',f"dch04a: type has too many / {docID}")
      if len(docType.split('/'))==0:
        reply+= outputString(outputStyle,'unsure',f"dch04b: no type in (removed data?) {docID}")
        continue
      if docType.startswith('x') and not docType.startswith(('x0','x1')):
        errorStr = outputString(outputStyle,'error',f"dch04c: bad data type*: {docID} {docType}")
        if repair is None:
          reply+= errorStr
        elif repair(errorStr):
          self.cursor.execute(f"UPDATE main SET type='x1' WHERE id = '{docID}'")
      if not all(k.startswith('x-') for k in stack.split('/')[:-1]):
        reply+= outputString(outputStyle,'error',f"dch03: non-text in stack in id: {docID}")
      if any(len(i)==0 for i in stack) and not docType.startswith('x0'): #if no inheritance
        if docType.startswith(('measurement','x')):
          reply+= outputString(outputStyle,'warning',f"branch stack length = 0: no parent {docID}")
        elif not minimal:
          reply+= outputString(outputStyle,'ok',
                                   f"branch stack length = 0: no parent for procedure/sample {docID}")
      try:
        dirNamePrefix = path.split(os.sep)[-1].split('_')[0]
        if dirNamePrefix.isdigit() and child!=int(dirNamePrefix) and docType.startswith('x'): #compare child-number to start of directory name
          reply+= outputString(outputStyle,'error',f"dch05: child-number and dirName dont match {docID}")
      except Exception:
        pass  #handled next lines
      if path is None:
        if docType.startswith('x'):
          reply+= outputString(outputStyle,'error',f"dch06: branch path is None {docID}")
        elif docType.startswith('measurement'):
          if not minimal:
            reply+= outputString(outputStyle,'ok', f'measurement branch path is None=no data {docID}')
        elif not minimal:
          reply+= outputString(outputStyle,'perfect',f"procedure/sample with empty path {docID}")
      else:                                                    #if sensible path
        if len(stack.split('/')) != len(path.split(os.sep)) and path!='*' and not path.startswith('http'):
          #check if length of path and stack coincide; ignore path=None=*
          if docType.startswith('procedure'):
            if not minimal:
              reply+= outputString(outputStyle,'perfect',
                                       f"procedure: branch stack and path lengths not equal: {docID}: {name}")
          else:
            reply+= outputString(outputStyle,'unsure',
              f"branch stack and path lengths not equal: {docID}: {name} {len(stack.split('/'))}"
              f" {len(path.split(os.sep))}")
        if path!='*' and not path.startswith('http'):
          for parentID in stack.split('/')[:-1]:            #check if all parents in doc have a corresponding path
            parentDoc = self.getDoc(parentID)
            if not parentDoc:
              errorStr= outputString(outputStyle,'error',f"branch stack parent is bad: {docID}")
              if repair is None:
                reply+= errorStr
              elif repair(errorStr):
                pathNew = (lostAndFoundProjPath/createDirName(name, 'x0', 0)).as_posix() if docID.startswith('x-')\
                          else '*'
                stackNew = f'{lostAndFoundProjId}/{docID}'
                self.cursor.execute(f"UPDATE branches SET path='{pathNew}', stack='{stackNew}' "\
                                    f"WHERE id='{docID}' AND stack='{stack}'")
                self.connection.commit()
              continue
            parentDocBranches = parentDoc['branch']
            onePathFound = any(path.startswith(parentBranch['path']) for parentBranch in parentDocBranches)
            if not onePathFound:
              if docType.startswith('procedure'):
                if not minimal:
                  reply+= outputString(outputStyle,'perfect',
                    f"dch08: procedure parent does not have corresponding path {docID} | parentID {parentID}")
              else:
                reply+= outputString(outputStyle,'unsure',
                    f"dch08: parent does not have corresponding path {docID} | parentID {parentID}")

    self.cursor.execute('SELECT id, idx FROM branches WHERE idx<0')
    for docID, idx in self.cursor.fetchall():
      reply+= outputString(outputStyle,'error',f"branch idx is bad: {docID} idx: {idx}")

    self.cursor.execute("SELECT id, key FROM properties where key NOT LIKE '%.%'")
    for docID, key in self.cursor.fetchall():
      reply+= outputString(outputStyle,'error',f"key is bad, miss .: {docID} idx: {key}")
    self.cursor.execute("SELECT id, key FROM properties where value LIKE ''")
    for docID, key in self.cursor.fetchall():
      reply+= outputString(outputStyle,'ok',f"value of this key is missing: {docID} idx: {key}")

    #doc-type specific tests
    cmd = "SELECT qrCodes.id, qrCodes.qrCode FROM qrCodes JOIN main USING(id) WHERE  main.type LIKE 'sample%'"
    self.cursor.execute(cmd)
    if res:= [i[0] for i in self.cursor.fetchall() if i[1] is None]:
      reply+= outputString(outputStyle,'warning',f"dch09: qrCode not in samples {res}")
    self.cursor.execute("SELECT id, shasum, image, comment FROM main WHERE  type LIKE 'measurement%'")
    for row in self.cursor.fetchall():
      docID, shasum, image, comment = row
      if shasum is None:
        reply+= outputString(outputStyle,'warning',f"dch10: shasum not in measurement {docID}")
      if image.startswith('data:image'):  #for jpg and png
        try:
          imgData = base64.b64decode(image[22:])
          Image.open(io.BytesIO(imgData))  #can convert, that is all that needs to be tested
        except Exception:
          reply+= outputString(outputStyle,'error',f"dch12: jpg-image not valid {docID}")
      elif image.startswith('<?xml'):
        #from https://stackoverflow.com/questions/63419010/check-if-an-image-file-is-a-valid-svg-file-in-python
        SVG_R = r'(?:<\?xml\b[^>]*>[^<]*)?(?:<!--.*?-->[^<]*)*(?:<svg|<!DOCTYPE svg)\b'
        SVG_RE = re.compile(SVG_R, re.DOTALL)
        if SVG_RE.match(image) is None:
          reply+= outputString(outputStyle,'error',f"dch13: svg-image not valid {docID}")
      elif image in ('', None):
        comment = comment.replace('\n','..')
        reply+=outputString(outputStyle,'unsure',f"image does not exist {docID} image:{image} comment:{comment}")
      else:
        reply+= outputString(outputStyle,'error',f"dch14: image not valid {docID} {image}")

    #test hierarchy
    for projID in self.getView('viewDocType/x0')['id'].values:
      _, error = self.getHierarchy(projID, True)
      if error:
        reply+= outputString(outputStyle,'error',f"dch15: project hierarchy invalid in project {projID}")
    if repair is not None:
      self.connection.commit()
    return reply



""" Adapter for converting between ORM models and domain models. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: orm_model_adapter.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

from pasta_eln.database.models.config_model import ConfigModel
from pasta_eln.database.models.config_orm_model import ConfigOrmModel
from pasta_eln.database.models.data_hierarchy_definition_model import DataHierarchyDefinitionModel
from pasta_eln.database.models.data_hierarchy_definition_orm_model import DataHierarchyDefinitionOrmModel
from pasta_eln.database.models.data_hierarchy_model import DataHierarchyModel
from pasta_eln.database.models.data_hierarchy_orm_model import DataHierarchyOrmModel
from pasta_eln.database.models.main_orm_model import MainOrmModel
from pasta_eln.database.models.project_model import ProjectModel
from pasta_eln.database.models.upload_model import UploadModel
from pasta_eln.database.models.upload_orm_model import UploadOrmModel


class OrmModelAdapter:
  """Adapter for converting between ORM models and domain models.

  This class provides static methods to convert various domain models to their
  corresponding ORM representations and vice versa. It facilitates the interaction
  between the application layer and the database layer by handling the necessary
  transformations.
  """

  @staticmethod
  def get_orm_config_model(model: ConfigModel) -> ConfigOrmModel:
    """Converts a ConfigModel to a DatabaseOrmConfigModel.

    This method takes a ConfigModel instance and transforms it into a
    DatabaseOrmConfigModel instance, adjusting the necessary fields.

    Args:
        model (ConfigModel): The configuration model to convert.

    Returns:
        ConfigOrmModel: The converted ORM configuration model.
    """
    model_dict = dict(model)
    model_dict['metadata_info'] = model_dict.pop('metadata')
    return ConfigOrmModel(**model_dict)

  @staticmethod
  def get_orm_upload_model(model: UploadModel) -> UploadOrmModel:
    """Converts an UploadModel to a DatabaseOrmUploadModel.

    This method takes an UploadModel instance and transforms it into a
    DatabaseOrmUploadModel instance.

    Args:
        model (UploadModel): The upload model to convert.

    Returns:
        UploadOrmModel: The converted ORM upload model.
    """
    model_dict = dict(model)
    return UploadOrmModel(**model_dict)

  @staticmethod
  def get_orm_project_model(model: ProjectModel) -> MainOrmModel:
    """Converts a ProjectModel to a DatabaseOrmMainModel.

    This method takes a ProjectModel instance and transforms it into a
    DatabaseOrmMainModel instance, adjusting the necessary fields.

    Args:
        model (ProjectModel): The project model to convert.

    Returns:
        MainOrmModel: The converted ORM project model.
    """
    model_dict = dict(model)
    model_dict.pop('status')
    model_dict.pop('objective')
    return MainOrmModel(**model_dict)

  @staticmethod
  def get_orm_data_hierarchy_model(model: DataHierarchyModel) -> DataHierarchyOrmModel:
    """Converts a DataHierarchyModel to a DatabaseOrmDataHierarchyModel.

    This method takes a DataHierarchyModel instance and transforms it into a
    DatabaseOrmDataHierarchyModel instance.

    Args:
        model (DataHierarchyModel): The data hierarchy model to convert.

    Returns:
        DataHierarchyOrmModel: The converted ORM data hierarchy model.
    """
    model_dict = dict(model)
    model_dict.pop('id')
    definitions = model_dict.pop('definitions')
    model_dict['definitions'] = [OrmModelAdapter.get_data_hierarchy_definition_orm_model(definition) for definition
                                 in definitions] if definitions else []
    return DataHierarchyOrmModel(**model_dict)

  @staticmethod
  def get_config_model(model: ConfigOrmModel) -> ConfigModel:
    """Converts a DatabaseOrmConfigModel to a ConfigModel.

    This method takes a DatabaseOrmConfigModel instance and transforms it into a
    ConfigModel instance, adjusting the necessary fields.

    Args:
        model (ConfigOrmModel): The ORM configuration model to convert.

    Returns:
        ConfigModel: The converted configuration model.
    """
    model_dict = dict(model)
    model_dict['metadata'] = model_dict.pop('metadata_info')
    model_dict['_id'] = model_dict.pop('id')
    return ConfigModel(**model_dict)

  @staticmethod
  def get_upload_model(model: UploadOrmModel) -> UploadModel:
    """Converts a DatabaseOrmUploadModel to an UploadModel.

    This method takes a DatabaseOrmUploadModel instance and transforms it into an
    UploadModel instance.

    Args:
        model (UploadOrmModel): The ORM upload model to convert.

    Returns:
        UploadModel: The converted upload model.
    """
    model_dict = dict(model)
    model_dict['_id'] = model_dict.pop('id')
    return UploadModel(**model_dict)

  @staticmethod
  def get_project_model(model: tuple[MainOrmModel, str, str]) -> ProjectModel:
    """Converts a tuple of DatabaseOrmMainModel and additional fields (status, objective) to a ProjectModel.

    This method takes a tuple containing a DatabaseOrmMainModel instance and
    additional fields, transforming it into a ProjectModel instance.

    Args:
        model (Tuple[MainOrmModel, str, str]): The tuple containing the ORM
        main model and additional fields.

    Returns:
        ProjectModel: The converted project model.
    """
    main_model_dict = dict(model[0])
    main_model_dict['_id'] = main_model_dict.pop('id')
    main_model_dict.pop('type')
    return ProjectModel(**main_model_dict, status=model[1], objective=model[2])

  @staticmethod
  def get_data_hierarchy_model(model: DataHierarchyOrmModel) -> DataHierarchyModel:
    """Converts a DatabaseOrmDataHierarchyModel to a DataHierarchyModel.

    This method takes a DatabaseOrmDataHierarchyModel instance and transforms it
    into a DataHierarchyModel instance.

    Args:
        model (DataHierarchyOrmModel): The ORM data hierarchy model to convert.

    Returns:
        DataHierarchyModel: The converted data hierarchy model.
    """
    model_dict = dict(model)
    definitions = model_dict.pop('definitions')
    model_dict['definitions'] = [OrmModelAdapter.get_data_hierarchy_definition_model(definition) for definition in
                                 definitions]
    return DataHierarchyModel(**model_dict)

  @staticmethod
  def get_data_hierarchy_definition_model(model: DataHierarchyDefinitionOrmModel) -> DataHierarchyDefinitionModel:
    """Converts a DatabaseOrmDataHierarchyDefinitionsModel to a DataHierarchyDefinitionsModel.

    This method takes a DatabaseOrmDataHierarchyDefinitionsModel instance and transforms it
    into a DataHierarchyDefinitionsModel instance.

    Args:
        model (DatabaseOrmDataHierarchyDefinitionsModel): The ORM data hierarchy definitions model to convert.

    Returns:
        DataHierarchyDefinitionModel: The converted data hierarchy definitions model.
    """
    model_dict = dict(model)
    meta_list: str = model_dict.pop('meta_list')
    model_dict['meta_list'] = meta_list.split(',') if meta_list else []
    return DataHierarchyDefinitionModel(**model_dict)

  @staticmethod
  def get_data_hierarchy_definition_orm_model(model: DataHierarchyDefinitionModel) -> DataHierarchyDefinitionOrmModel:
    """
    Convert a DataHierarchyDefinitionModel instance to a DataHierarchyDefinitionOrmModel instance.

    This static method takes a DataHierarchyDefinitionModel and transforms it into a
    DataHierarchyDefinitionOrmModel by converting the meta_list into a comma-separated
    string and excluding the 'id' field from the resulting model.

    Args:
        model (DataHierarchyDefinitionModel): The data hierarchy definition model to convert.

    Returns:
        DataHierarchyDefinitionOrmModel: The converted ORM data hierarchy definition model.

    """

    model_dict = dict(model)
    model_dict.pop('id')
    meta_list = model_dict.pop('meta_list')
    model_dict['meta_list'] = ','.join(meta_list) if meta_list and isinstance(meta_list, list) else ''
    return DataHierarchyDefinitionOrmModel(**model_dict)



""" Functions that modify strings into the appropriate format """
import logging
import re
import traceback
from ..miscTools import Bcolors


def outputString(fmt:str='print', level:str='info', message:str='') -> str:
  """ Output a message into different formats:
    - print: print to stdout
    - logging; log to file
    - text: return text string (supersedes html)
    - html: return html string https://doc.qt.io/qtforpython/overviews/richtext-html-subset.html#supported-html-subset
    - else: no output
    - formats can be union ('print,text')
  """
  prefixes = {'h2':f'{Bcolors.UNDERLINE}\n*** ','bold':f'{Bcolors.BOLD}\n*** ', \
              'perfect':f'{Bcolors.OKGREEN}', 'ok':f'{Bcolors.OKBLUE}', 'unsure':f'{Bcolors.HEADER}',\
              'warning':f'{Bcolors.WARNING}**Warning','error':f'{Bcolors.FAIL}**ERROR '}
  if level=='info':
    txtOutput = message.strip()+'\n'
  elif level in prefixes:
    txtOutput = prefixes[level]+message
    txtOutput+= ' ***' if '***' in prefixes[level] else ''
    txtOutput+= f'{Bcolors.ENDC}\n'
  else:
    print('ERROR level not in prefixes ',level)
  # depend on format
  if 'print' in fmt:
    print(txtOutput)
  if 'logging' in fmt and level in {'info', 'warning', 'error'}:
    getattr(logging,level)(message)
  if 'text' in fmt:
    return txtOutput
  if fmt=='html':
    colors = {'info':'black','error':'red','warning':'orangered','perfect':'green','ok':'blue','unsure':'darkmagenta'}
    if level[0]=='h':
      return f'<{level}>{message}</{level}>'
    if level not in colors:
      print(f'**ERROR: wrong level {level}')
      return ''
    return f'<font color="{colors[level]}">' + message.replace('\n', '<br>') + '</font><br>'
  return ''


def tracebackString(log:bool=False, docID:str='') -> str:
  """ Create a formatted string of traceback

  Args:
    log (bool): write to logging
    docID (str): docID used in comment

  Returns:
    str: | separated string of call functions
  """
  tracebackList = [i.split('\n')[0] for i in traceback.format_stack()[2:-2] if 'pasta_eln' in i] #skip first and last and then filter only things with pasta_eln
  reply = '|'.join([item.split('/')[-1].strip() for item in tracebackList])  #| separated list of stack excluding last
  reply = reply.replace('",','')
  if log:
    logging.info(' traceback %s %s', docID, reply)
  return reply


def markdownEqualizer(text:str) -> str:
  """
  Create a markdown that well balanced with regard to font size, etc.

  Args:
    text (str): input string

  Returns:
    str: output str
  """
  if isinstance(text, tuple):
    text=text[0]
  return re.sub(r'(^|\n)(#+)', r'\1##\2', text.strip())


def camelCase(text:str) -> str:
  """
  Produce camelCase from normal string
  - file names abcdefg.hij are only replaced spaces

  Args:
     text (str): string

  Returns:
    str: camel case of that string: CamelCaseString
  """
  if re.match(r'^[\w-]+\.[\w]+$', text):
    return text.replace(' ','_')
  return re.sub(r'(_|-)+', ' ', text).title().replace(' ','').replace('*','')


def createDirName(name:str, docType:str, thisChildNumber:int) -> str:
  """ create directory-name by using camelCase and a prefix

  Args:
      name (string): name with spaces etc.
      docType (string): document type used for prefix
      thisChildNumber (int): number of myself

  Returns:
    string: directory name with leading number
  """
  if docType == 'x0':
    return camelCase(name)
  #steps, tasks
  return f'{thisChildNumber:03d}_{camelCase(name)}'



""" Data type class names. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: data_type_class_names.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

from enum import Enum


class DataTypeClassName(Enum):
  """
  An enumeration representing different data type class names.

  Explanation:
      This enumeration defines the names of different data type classes such as primitive, compound, and controlled vocabulary.

  """
  PRIMITIVE = 'primitive'
  COMPOUND = 'compound'
  CONTROLLED_VOCAB = 'controlledVocabulary'

  def __str__(self) -> str:
    """
    Returns the string representation of the data type class name.

    Explanation:
        This method returns the string value of the data type class name.

    Returns:
        str: The string representation of the data type class name.
    """
    return self.value



""" Widget that shows the content of project in a electronic labnotebook """
import importlib
import logging
from enum import Enum
from pathlib import Path
from typing import Any, Optional
from anytree import Node, PreOrderIter
from PySide6.QtCore import QItemSelectionModel, QModelIndex, Qt, Slot  # pylint: disable=no-name-in-module
from PySide6.QtGui import QAction, QStandardItem, QStandardItemModel  # pylint: disable=no-name-in-module
from PySide6.QtWidgets import (QLabel, QMenu, QMessageBox, QTextEdit, QVBoxLayout,  # pylint: disable=no-name-in-module
                               QWidget)
from ..fixedStringsJson import DO_NOT_RENDER
from ..guiCommunicate import Communicate
from ..guiStyle import Action, Label, TextButton, showMessage, widgetAndLayout
from ..textTools.handleDictionaries import doc2markdown
from ..textTools.stringChanges import createDirName
from .projectTreeView import TreeView


class Project(QWidget):
  """ Widget that shows the content of project in a electronic labnotebook """
  def __init__(self, comm:Communicate):
    super().__init__()
    self.comm = comm
    comm.changeProject.connect(self.change)
    self.mainL = QVBoxLayout()
    self.setLayout(self.mainL)
    self.tree:Optional[TreeView]             = None
    self.model:Optional[QStandardItemModel]  = None
    self.allDetails                          = QTextEdit()
    self.actHideDetail                       = QAction()
    self.actionFoldAll                       = QAction()
    self.projID = ''
    self.docProj:dict[str,Any]= {}
    self.showAll= True
    self.showDetailsAll = False
    self.btnAddSubfolder:Optional[TextButton] = None
    self.lineSep = 20


  def projHeader(self) -> None:
    """
    Initialize / Create header of page
    """
    self.docProj = self.comm.backend.db.getDoc(self.projID)
    dataHierarchyNode = self.comm.backend.db.dataHierarchy('x0', 'meta')
    # remove if still there
    for i in reversed(range(self.mainL.count())): #remove old
      self.mainL.itemAt(i).widget().setParent(None)
    logging.debug('ProjectView elements at 2: %i',self.mainL.count())
    # TOP LINE includes name on left, buttons on right
    _, topLineL       = widgetAndLayout('H',self.mainL,'m')
    hidden, menuTextHidden = ('     \U0001F441', 'Mark project as shown') \
                       if [b for b in self.docProj['branch'] if False in b['show']] else \
                       ('', 'Mark project as hidden')
    topLineL.addWidget(Label(self.docProj['name']+hidden, 'h2'))
    showStatus = '(Show all items)' if self.showAll else '(Hide hidden items)'
    topLineL.addWidget(QLabel(showStatus))
    topLineL.addStretch(1)
    # buttons in top line
    buttonW, buttonL = widgetAndLayout('H', spacing='m')
    topLineL.addWidget(buttonW, alignment=Qt.AlignTop)  # type: ignore
    self.btnAddSubfolder = TextButton('Add subfolder', self, [Command.ADD_CHILD], buttonL)
    TextButton('Edit project',                         self, [Command.EDIT],      buttonL)
    visibility = TextButton(          'Visibility',    self, [],                  buttonL)
    visibilityMenu = QMenu(self)
    self.actHideDetail = Action('Hide project details',self, [Command.SHOW_PROJ_DETAILS],visibilityMenu)
    menuTextItems = 'Hide hidden items' if self.showAll else 'Show hidden items'
    minimizeItems = 'Show all item details' if self.showDetailsAll else 'Hide all item details'
    Action( menuTextItems,    self, [Command.HIDE_SHOW_ITEMS],  visibilityMenu)
    Action( menuTextHidden,   self, [Command.HIDE],             visibilityMenu)
    self.actionFoldAll     = Action( minimizeItems,    self, [Command.SHOW_DETAILS],     visibilityMenu)
    visibility.setMenu(visibilityMenu)
    more = TextButton('More',           self, [], buttonL)
    moreMenu = QMenu(self)
    Action('Scan',                      self, [Command.SCAN], moreMenu)
    for doctype in self.comm.backend.db.dataHierarchy('', ''):
      if doctype[0]!='x':
        icon = self.comm.backend.db.dataHierarchy(doctype, 'icon')[0]
        icon = 'fa5s.asterisk' if icon=='' else icon
        Action(f'table of {doctype}',   self, [Command.SHOW_TABLE, doctype], moreMenu, icon=icon)
    Action('table of unidentified',     self, [Command.SHOW_TABLE, '-'],     moreMenu, icon='fa5.file')
    moreMenu.addSeparator()
    projectGroup = self.comm.backend.configuration['projectGroups'][self.comm.backend.configurationProjectGroup]
    if projectAddOns := projectGroup.get('addOns',{}).get('project',''):
      for label, description in projectAddOns.items():
        Action(description, self, [Command.ADD_ON, label], moreMenu)
      moreMenu.addSeparator()
    Action('Delete project',            self, [Command.DELETE], moreMenu)
    more.setMenu(moreMenu)

    # Details section
    # self.infoW = QScrollArea()
    # self.infoW.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
    # self.infoW.setWidgetResizable(True)
    self.allDetails.setMarkdown(doc2markdown(self.docProj, DO_NOT_RENDER, dataHierarchyNode, self))
    if not self.docProj['gui'][0]:
      self.allDetails.hide()
      self.actHideDetail.setText('Show project details')
    self.allDetails.resizeEvent = self.commentResize # type: ignore
    bgColor = self.comm.palette.get('secondaryDark', 'background-color')
    fgColor = self.comm.palette.get('secondaryText', 'color')
    self.allDetails.setStyleSheet(f"border: none; padding: 0px; {bgColor} {fgColor}")
    self.allDetails.setReadOnly(True)
    self.mainL.addWidget(self.allDetails)
    self.commentResize(None)
    return


  def commentResize(self, _:Any) -> None:
    """ called if comment is resized because widget initially/finally knows its size
    - comment widget is hard coded size it depends on the rendered size
    """
    if self.allDetails is None:
      return
    self.allDetails.document().setTextWidth(self.width())
    height:int = self.allDetails.document().size().toTuple()[1]  # type: ignore[index]
    self.allDetails.setMaximumHeight(height+12)
    return


  @Slot(str, str)                                   # type: ignore[arg-type]
  def change(self, projID:str, docID:str) -> None:
    """
    What happens when user clicks to change project

    Args:
      projID (str): document id of project; if empty, just refresh
      docID (str): document id of focus item, if not given focus at project
    """
    logging.debug('project:changeProject |%s|%s|',projID,docID)
    #initialize
    for i in reversed(range(self.mainL.count())): #remove old
      self.mainL.itemAt(i).widget().setParent(None)
    logging.debug('ProjectView elements at 1: %i',self.mainL.count())
    if projID!='':
      self.projID         = projID
      self.comm.projectID = projID
    selectedIndex = None
    self.model = QStandardItemModel()
    self.tree = TreeView(self, self.comm, self.model)
    # self.tree.setSelectionBehavior(QAbstractItemView.SelectRows)
    # self.tree.setSelectionMode(QAbstractItemView.SingleSelection)
    self.model.itemChanged.connect(self.modelChanged)
    rootItem = self.model.invisibleRootItem()
    #Populate model body of change project: start recursion
    nodeHier, error = self.comm.backend.db.getHierarchy(self.projID, allItems=self.showAll)
    if error:
      showMessage(self, 'Error', 'There is an error in the project hierarchy: a parent of a node is incorrect.', 'Critical')
    for node in PreOrderIter(nodeHier, maxlevel=2):
      if node.is_root:         #Project header
        self.projHeader()
      else:
        rootItem.appendRow(self.iterateTree(node))
    # collapse / expand depending on stored value
    # by iterating each leaf, and converting item and index
    root = self.model.invisibleRootItem()
    self.setExpandedState(root)
    if selectedIndex is not None:
      self.tree.selectionModel().select(selectedIndex, QItemSelectionModel.Select)
      self.tree.setCurrentIndex(selectedIndex)# Item(selectedItem)
    self.mainL.addWidget(self.tree)
    logging.debug('ProjectView elements at 4: %i',self.mainL.count())
    if len(nodeHier.children)>0 and self.btnAddSubfolder is not None:
      self.btnAddSubfolder.setVisible(False)
    self.tree.expanded.connect(lambda index: self.actionExpandCollapse(index, True))
    self.tree.collapsed.connect(lambda index: self.actionExpandCollapse(index, False))
    return


  def setExpandedState(self, node:QStandardItem) -> None:
    """ Recursive function to set the expanded state of nodes

    Args:
      node (QStandardItem): node to process
    """
    if self.model is None or self.tree is None:
      return
    for iRow in range(node.rowCount()):
      item = node.child(iRow)
      data = item.data(role=Qt.ItemDataRole.UserRole+1)
      if data['hierStack'].split('/')[-1][0]=='x':
        index = self.model.indexFromItem(item)
        self.tree.setExpanded(index, data['gui'][1])
      self.setExpandedState(item)
    return


  def actionExpandCollapse(self, index:QModelIndex, flag:bool) -> None:
    """ Action upon expansion or collapsing of folder (showing its children)

    Args:
      index (QModelIndex): index that send the signal
      flag (bool): true=expand=show-children, false=collapse=hide-children
    """
    if self.model is None:
      return
    gui  = [index.data(Qt.ItemDataRole.UserRole+1)['gui'][0]]+[flag]
    docID = index.data(Qt.ItemDataRole.UserRole+1)['hierStack'].split('/')[-1]
    self.model.itemFromIndex(index).setData({ **index.data(Qt.ItemDataRole.UserRole+1), **{'gui':gui}})
    self.comm.backend.db.setGUI(docID, gui)
    return


  def execute(self, command:list[Any]) -> None:
    """
    Event if user clicks button in the center

    Args:
      command (list): list of commands
    """
    if command[0] is Command.EDIT:
      self.comm.formDoc.emit(self.docProj)
      self.change(self.projID,'')
      #collect information and then change
      oldPath = self.comm.backend.basePath/self.docProj['branch'][0]['path']
      if oldPath.is_dir():
        newPath = self.comm.backend.basePath/createDirName(self.docProj['name'],'x0',0)
        oldPath.rename(newPath)
      self.comm.changeSidebar.emit('redraw')
    elif command[0] is Command.DELETE:
      ret = QMessageBox.critical(self, 'Warning', 'Are you sure you want to delete project?', \
                      QMessageBox.StandardButton.No | QMessageBox.StandardButton.Yes,
                      QMessageBox.StandardButton.No)
      if ret==QMessageBox.StandardButton.Yes:
        #delete database and rename folder
        doc = self.comm.backend.db.remove(self.projID)
        if 'branch' in doc and len(doc['branch'])>0 and 'path' in doc['branch'][0]:
          oldPath = self.comm.backend.basePath/doc['branch'][0]['path']
          newPath = self.comm.backend.basePath/('trash_'+doc['branch'][0]['path'])
          nextIteration = 1
          while newPath.is_dir():
            newPath = self.comm.backend.basePath/(f"trash_{doc['branch'][0]['path']}_{nextIteration}")
            nextIteration += 1
          oldPath.rename(newPath)
        # go through children, remove from DB
        children = self.comm.backend.db.getView('viewHierarchy/viewHierarchy', startKey=self.projID)
        for line in children:
          self.comm.backend.db.remove(line['id'])
        #update sidebar, show projects
        self.comm.changeSidebar.emit('redraw')
        self.comm.changeTable.emit('x0','')
    elif command[0] is Command.SCAN:
      for _ in range(2):  #scan twice: convert, extract
        self.comm.backend.scanProject(None, self.projID, self.docProj['branch'][0]['path'])
      self.comm.changeProject.emit(self.projID,'')
      showMessage(self, 'Information','Scanning finished')
    elif command[0] is Command.SHOW_PROJ_DETAILS:
      self.docProj['gui'][0] = not self.docProj['gui'][0]
      self.comm.backend.db.setGUI(self.projID, self.docProj['gui'])
      if self.allDetails is not None and self.allDetails.isHidden():
        self.allDetails.show()
        self.actHideDetail.setText('Hide project details')
      elif self.allDetails is not None:
        self.allDetails.hide()
        self.actHideDetail.setText('Show project details')
    elif command[0] is Command.HIDE:
      self.comm.backend.db.hideShow(self.projID)
      self.docProj = self.comm.backend.db.getDoc(self.projID)
      if [b for b in self.docProj['branch'] if False in b['show']]: # hidden->go back to project table
        self.comm.changeSidebar.emit('')
        self.comm.changeTable.emit('x0','') # go back to project table
      else:
        self.change('', '')
        self.comm.changeSidebar.emit('')
    elif command[0] is Command.SHOW_DETAILS and self.tree is not None:
      def recursiveRowIteration(index:QModelIndex) -> None:
        for subRow in range(self.tree.model().rowCount(index)):   # type: ignore[union-attr]
          subIndex = self.tree.model().index(subRow,0, index)     # type: ignore[union-attr]
          subItem  = self.tree.model().itemFromIndex(subIndex)    # type: ignore[union-attr]
          docID    = subItem.data()['hierStack'].split('/')[-1]
          gui      = subItem.data()['gui']
          gui[0]   = self.showDetailsAll
          subItem.setData({ **subItem.data(), **{'gui':gui}})
          self.comm.backend.db.setGUI(docID, gui)
          recursiveRowIteration(subIndex)
      recursiveRowIteration(self.tree.model().index(-1,0))
      self.showDetailsAll = not self.showDetailsAll
      if self.showDetailsAll:
        self.actionFoldAll.setText('Show all item details')
      else:
        self.actionFoldAll.setText('Hide all item details')
    elif command[0] is Command.HIDE_SHOW_ITEMS:
      self.showAll = not self.showAll
      self.change('','')
    elif command[0] is Command.ADD_CHILD:
      self.comm.backend.cwd = self.comm.backend.basePath/self.docProj['branch'][0]['path']
      self.comm.backend.addData('x1', {'name':'new item'}, [self.projID])
      self.change('','') #refresh project
    elif command[0] is Command.SHOW_TABLE:
      self.comm.changeTable.emit(command[1], self.projID)
    elif command[0] is Command.ADD_ON:
      module      = importlib.import_module(command[1])
      module.main(self.comm.backend, self.projID, self)
    else:
      print('**ERROR project menu unknown:',command)
    return


  def modelChanged(self, item:QStandardItem) -> None:
    """
    Autocalled after drag-drop and other changes, record changes to backend and database directly

    Args:
      item (QStandardItem): item changed, new location
    """
    #gather old information
    db       = self.comm.backend.db
    if not item.data():
      return
    stackOld = item.data()['hierStack'].split('/')[:-1]
    docID    = item.data()['hierStack'].split('/')[-1]
    doc      = db.getDoc(docID)
    if 'branch' not in doc or not stackOld: #skip everything if project or not contain branch
      return
    branchOldList= [i for i in doc['branch'] if i['stack']==stackOld]
    if len(branchOldList)!=1:
      self.change('','')
      return
    branchOld = branchOldList[0]
    childOld = branchOld['child']
    siblingsOld = db.getView('viewHierarchy/viewHierarchy', startKey='/'.join(stackOld))
    siblingsOld = [i for i in siblingsOld if len(i['key'].split('/'))==len(stackOld)+1 and \
                                                  i['value'][0]>branchOld['child'] and i['value'][0]<9999]
    #gather new information
    stackNew = []  #create reversed
    currentItem = item
    while currentItem.parent() is not None:
      currentItem = currentItem.parent()
      docIDj = currentItem.data()['hierStack'].split('/')[-1]
      stackNew.append(docIDj)
    stackNew = [self.projID] + stackNew[::-1]  #add project id and reverse
    childNew = item.row()
    if branchOld['path'] is not None and not branchOld['path'].startswith('http'):
      if doc['type'][0][0]=='x':
        dirNameNew= createDirName(doc['name'],doc['type'][0],childNew) # create path name: do not create directory on storage yet
      else:
        dirNameNew= Path(branchOld['path']).name                         # use old name
      parentDir = db.getDoc(stackNew[-1])['branch'][0]['path']
      pathNew = f'{parentDir}/{dirNameNew}'
    else:
      pathNew = branchOld['path']
    siblingsNew = db.getView('viewHierarchy/viewHierarchy', startKey='/'.join(stackNew))
    siblingsNew = [i for i in siblingsNew if len(i['key'].split('/'))==len(stackNew)+1 and \
                                                   i['value'][0]>=childNew and i['value'][0]<9999]
    logging.debug('Change project: docID %s | old stack %s child %i | new stack %s child %i path %s'\
                  , docID, str(stackOld), childOld, str(stackNew), childNew, pathNew)
    if stackOld==stackNew and childOld==childNew:  #nothing changed, just redraw
      return
    # CHANGE
    # change new siblings
    for line in siblingsNew:
      if line['id']!=docID:
        db.updateBranch(docID=line['id'], branch=line['value'][4], child=line['value'][0]+1)
    # change item in question
    db.updateBranch(docID=docID, branch=-99, stack=stackNew, path=pathNew, child=childNew, stackOld=stackOld+[docID])
    item.setData(item.data() | {'hierStack': '/'.join(stackNew+[docID])})
    # change old siblings
    for line in siblingsOld:
      db.updateBranch(  docID=line['id'], branch=line['value'][4], child=line['value'][0]-1)
    return


  def iterateTree(self, nodeHier:Node) -> QStandardItem:
    """
    Recursive function to translate the hierarchical node into a tree-node

    Args:
      nodeHier (Anytree.Node): anytree node

    Returns:
      QtTreeWidgetItem: tree node
    """
    #prefill docID
    hierStack = '/'.join([i.id for i in nodeHier.ancestors]+[nodeHier.id])
    gui = nodeHier.gui
    nodeTree = QStandardItem(nodeHier.name)
    nodeTree.setData({'hierStack':hierStack, 'docType':nodeHier.docType, 'gui':gui})
    if nodeHier.id[0]=='x':
      nodeTree.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsDragEnabled | Qt.ItemIsDropEnabled) # type: ignore
    else:
      nodeTree.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsDragEnabled) # type: ignore
    children = []
    for childHier in nodeHier.children:
      childTree = self.iterateTree(childHier)
      children.append(childTree)
    if children:
      nodeTree.appendRows(children)
    return nodeTree


class Command(Enum):
  """ Commands used in this file """
  EDIT   = 1
  DELETE = 2
  SCAN   = 3
  HIDE   = 4
  SHOW_PROJ_DETAILS = 5
  HIDE_SHOW_ITEMS   = 6
  SHOW_DETAILS      = 7
  ADD_CHILD         = 8
  SHOW_TABLE        = 9
  ADD_ON            = 10



""" PYTHON MIXIN FOR BACKEND containing all the functions that output to CLI """
# mypy: ignore-errors

class CLI_Mixin:
  """ Python Mixin for backend containing all the functions that output to CLI """

  def output(self, docType, printID=False):
    """
    output view to screen
    - length of output 100 character

    Args:
      docType (string): document type to output
      printID (bool):  include docID in output string

    Returns:
        string: output incl. \n
    """
    df = self.db.getView(f'viewDocType/{docType}')
    if not printID:
      df.drop('id', axis=1)
    output = df.to_string(index=False, justify='left', max_colwidth=40)
    outputList = output.split('\n')
    for colName in df.columns[1:]:
      idx = outputList[0].find(colName)
      outputList = [f'{i[:idx]}| {i[idx:]}' for i in outputList]
    outputList.insert(1, '-'*len(max(outputList, key = len)) )
    output = '\n'.join(outputList)
    return output


  def outputHierarchy(self, onlyHierarchy=True, addID=False):
    """
    output hierarchical structure in database
    - convert view into native dictionary
    - ignore key since it is always the same

    Args:
       onlyHierarchy (bool): only print project,steps,tasks or print all (incl. measurements...)[default print all]
       addID (bool): add docID to output

    Returns:
        string: output incl. \n
    """
    from anytree import PreOrderIter
    if len(self.hierStack) == 0:
      return 'Warning: pasta.outputHierarchy No project selected'
    hierString = ' '.join(self.hierStack)
    hierarchy, _ = self.db.getHierarchy(hierString)
    return ''.join('  '*node.depth + node.name + ' | ' + '/'.join(node.docType) + (f' | {node.id}' if addID else '') +'\n'
                   for node in PreOrderIter(hierarchy) if node.docType[0].startswith('x') or not onlyHierarchy)


  def outputQR(self):
    """
    output list of sample qr-codes

    Returns:
        string: output incl. \n
    """
    outString = f"{'QR': <36}|{'Name': <36}|{'ID': <36}\n"
    outString += '-'*110+'\n'
    for item in self.db.getView('viewIdentify/viewQR'):
      outString += f"{item['key'][:36]: <36}|{item['value'][:36]: <36}|{item['id'][:36]: <36}\n"
    return outString



""" Represents the ORM model for the data hierarchy table in the database. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: data_hierarchy_orm_model.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

from typing import Optional
from sqlalchemy.orm import Mapped, mapped_column, relationship
from pasta_eln.database.models.data_hierarchy_definition_orm_model import DataHierarchyDefinitionOrmModel
from pasta_eln.database.models.orm_model_base import OrmModelBase


class DataHierarchyOrmModel(OrmModelBase):
  """Represents the ORM model for the data hierarchy table in the database.

  This class defines the structure of the data hierarchy model used in the application,
  including fields for document type, IRI, title, icon, shortcut, and view. It provides
  a method to retrieve the names of the table columns for database operations.
  """
  __tablename__ = 'dataHierarchy'
  doc_type: Mapped[str] = mapped_column('docType', primary_key=True)
  IRI: Mapped[Optional[str]]
  title: Mapped[Optional[str]]
  icon: Mapped[Optional[str]]
  shortcut: Mapped[Optional[str]]
  view: Mapped[Optional[str]]
  definitions = relationship(
    DataHierarchyDefinitionOrmModel,
    primaryjoin='DataHierarchyDefinitionOrmModel.doc_type==DataHierarchyOrmModel.doc_type',
    foreign_keys='DataHierarchyDefinitionOrmModel.doc_type'
  )

  @classmethod
  def get_table_columns(cls) -> list[str]:
    """Retrieves the list of column names for the data hierarchy table.

    This method returns a list of strings representing the names of the columns
    defined in the data hierarchy model. It is useful for database operations that
    require knowledge of the table structure.

    Returns:
        list[str]: A list of column names for the data hierarchy table.
    """
    return ['doc_type', 'IRI', 'title', 'icon', 'shortcut', 'view', 'definitions']



""" ReorderColumnDelegate for the table views """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2023
#
#  Author: Jithu Murugan
#  Filename: reorder_column_delegate.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
from typing import Union
import qtawesome as qta
from pandas import DataFrame
from PySide6.QtCore import QAbstractItemModel, QEvent, QModelIndex, QPersistentModelIndex, QSize, Signal
from PySide6.QtGui import QPainter
from PySide6.QtWidgets import (QApplication, QPushButton, QStyle, QStyledItemDelegate, QStyleOptionButton,
                               QStyleOptionViewItem, QWidget)
from .utility_functions import is_click_within_bounds


class ReorderColumnDelegate(QStyledItemDelegate):
  """
  Delegate for creating the icons for the re-order column in the data hierarchy editor tables
  """
  re_order_signal = Signal(int)

  def __init__(self, df:DataFrame, group:str) -> None:
    """
      Constructor

    Args:
      df (DataFrame): pandas dataframe containing the entire schema
      group (str): string of this group/class
    """
    super().__init__()
    self.df = df
    self.group = group


  def paint(self,
            painter: QPainter,
            option: QStyleOptionViewItem,
            index: Union[QModelIndex, QPersistentModelIndex]) -> None:
    """
    Draws the re-order button within the cell represented by index
    Args:
      painter (QPainter): Painter instance for painting the button.
      option (QStyleOptionViewItem): Style option for the cell represented by index.
      index (Union[QModelIndex, QPersistentModelIndex]): Table cell index
    """
    if index.row()>=len(self.df[self.df['class']==self.group]):
      return
    button = QPushButton()
    opt = QStyleOptionButton()
    opt.state = QStyle.StateFlag.State_Active | QStyle.StateFlag.State_Enabled  # type: ignore[attr-defined]
    opt.rect = option.rect                                                      # type: ignore[attr-defined]
    opt.icon = qta.icon('fa5s.arrow-up', scale_factor=1.0)                      # type: ignore[attr-defined]
    opt.iconSize = QSize(15, 15)                                                # type: ignore[attr-defined]
    QApplication.style().drawControl(QStyle.ControlElement.CE_PushButton, opt, painter, button)
    return


  def createEditor(self,
                   parent: QWidget,
                   option: QStyleOptionViewItem,
                   index: Union[QModelIndex, QPersistentModelIndex]) -> QWidget:
    """
    Disable the editor for the whole re-order column by simply returning None
    Args:
      parent (QWidget): Parent table view.
      option (QStyleOptionViewItem): Style option for the cell represented by index.
      index (Union[QModelIndex, QPersistentModelIndex]): Cell index.
    """
    return None  # type: ignore[return-value]


  def editorEvent(self,
                  event: QEvent,
                  model: QAbstractItemModel,
                  option: QStyleOptionViewItem,
                  index: Union[QModelIndex, QPersistentModelIndex]) -> bool:
    """
    In case of mouse click event, the re_order_signal is emitted for the respective table cell position
    Args:
      event (QEvent): The editor event information.
      model (QAbstractItemModel): Model data representing the table view.
      option (QStyleOptionViewItem): QStyleOption for the table cell.
      index (Union[QModelIndex, QPersistentModelIndex]): Table cell index.

    Returns (bool): True/False
    """
    if is_click_within_bounds(event, option):
      self.re_order_signal.emit(index.row())
      return True
    return False



#!/usr/bin/python3
"""TEST using the FULL set of python-requirements: randomly move,copy,delete files to check resistance of scanning """
import os, shutil, logging, random
import warnings
import unittest
from pathlib import Path
from pasta_eln.backend import Backend
from pasta_eln.textTools.stringChanges import outputString
from pasta_eln.miscTools import DummyProgressBar
from pasta_eln.installationTools import exampleData

class TestStringMethods(unittest.TestCase):
  """
  derived class for this test
  """
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.be = None
    self.dirName = ''

  def test_main(self):
    """
    main function
    """
    outputFormat = ''  #change to 'print' for human usage, '' for less output
    dummyProgressBar = DummyProgressBar()
    # initialization: create database, destroy on filesystem and database and then create new one
    warnings.filterwarnings('ignore', message='numpy.ufunc size changed')
    warnings.filterwarnings('ignore', message='invalid escape sequence')
    warnings.filterwarnings('ignore', category=ResourceWarning, module='PIL')
    warnings.filterwarnings('ignore', category=ImportWarning)
    logPath = Path.home()/'pastaELN.log'
    logging.basicConfig(filename=logPath, level=logging.INFO, format='%(asctime)s|%(levelname)s:%(message)s',
                        datefmt='%m-%d %H:%M:%S')   #This logging is always info, since for installation only
    for package in ['urllib3', 'requests', 'asyncio', 'PIL', 'matplotlib.font_manager']:
      logging.getLogger(package).setLevel(logging.WARNING)
    logging.info('Start 02 test')
    self.be = Backend('research')
    projID = self.be.output('x0').split('|')[-1].strip()
    self.be.changeHierarchy(projID)

    progressBar = DummyProgressBar()
    choices = random.choices(range(100), k=155)
    # choices =  [34,83,40,8,70,1,14,14,60,42,13,99,94,47,14,7,45,85,96,80,91,35,33,85,85,38,63,1,89,62,24,47,82,1,13,4,97,46,6,82,97,76,5,14,69,7,50,14,33,8,15,69,99,75,69,68,9,70,56,75,44,60,97,93,72,50,46,89,60,63,65,82,50,54,33,6,77,52,72,90,82,1,58,11,81,76,9,3,83,40,26,45,88,56,10,74,66,16,5,31,27,62,65,65,56,7,80,43,56,15,93,36,49,74,15,73,5,80,35,46,17,31,79,51,20,81,5,50,49,31,59,9,28,9,33,65,81,85,10,78,64,79,58,14,99,35,86,98,95,0]
    print(f'Current choice: [{",".join([str(i) for i in choices])}]')
    for epoch in range(5):
      print(f'\nStart epoch: {epoch}')
      allDirs = []
      allFiles = []
      for root, _, files in os.walk(self.be.basePath):
        if root.endswith('StandardOperatingProcedures') or root==str(self.be.basePath):
          continue
        allDirs.append(root)
        allFiles += [f'{root}{os.sep}{i}' for i in files if i not in ('.id_pastaELN.json','pastaELN.db')]
      for iFile in allFiles:
        op = ['keep','copy','copy','move','move','delete'][choices.pop(0)%6]
        parent = Path(iFile).parent
        possTargets = [i for i in allDirs if Path(i)!=parent]
        target = possTargets[choices.pop(0)%len(possTargets)]
        if choices.pop(0)%4==0:
          target += f'/NewDirectory/{Path(iFile).name}'
        print(iFile[len(str(self.be.basePath))+1:], op, target[len(str(self.be.basePath))+1:])
        try:
          if op=='keep':
            continue
          elif op=='copy':
            if not Path(target).parent.is_dir():
              Path(target).parent.mkdir()
            shutil.copy(iFile, target)
          elif op=='move':
            if not Path(target).parent.is_dir():
              Path(target).parent.mkdir()
            shutil.move(iFile, target)
          elif op=='delete':
            os.unlink(iFile)
        except shutil.Error:
          pass  #skip step if invalid step because file already exists in that folder
      if choices.pop(0)%2==1:
        print('Start scanning')
        self.be.scanProject(progressBar, projID)
      if choices.pop(0)%2==1:
        print('Start scanning')
        self.be.scanProject(progressBar, projID)
    print('Final scanning and verification')
    self.be.scanProject(progressBar, projID)
    self.verify()
    return

  def verify(self):
    #Verify DB
    output = self.be.checkDB(outputStyle='text')
    output = '\n'.join(output.split('\n')[8:])
    print(output)
    self.assertNotIn('**ERROR', output, 'Error in checkDB')
    self.assertLessEqual(len(output.split('\n')), 6, 'Check db should have less than 6 almost empty lines')
    return

  def tearDown(self):
    logging.info('End 3Projects test')
    return

if __name__ == '__main__':
  unittest.main()



#!/usr/bin/python3
"""TEST the form """
import logging, warnings, shutil
from pathlib import Path
from PySide6.QtGui import QStandardItem
from pasta_eln.backend import Backend
from pasta_eln.miscTools import DummyProgressBar
from pasta_eln.installationTools import exampleData
from pasta_eln.GUI.form import Form
from pasta_eln.guiCommunicate import Communicate
from pasta_eln.GUI.palette import Palette
from pasta_eln.GUI.project import Project

def test_simple(qtbot):
  """
  main function
  """
  # initialization: create database, destroy on filesystem and database and then create new one
  warnings.filterwarnings('ignore', message='numpy.ufunc size changed')
  warnings.filterwarnings('ignore', message='invalid escape sequence')
  warnings.filterwarnings('ignore', category=ResourceWarning, module='PIL')
  warnings.filterwarnings('ignore', category=ImportWarning)
  logPath = Path.home()/'pastaELN.log'
  logging.basicConfig(filename=logPath, level=logging.INFO, format='%(asctime)s|%(levelname)s:%(message)s',
                      datefmt='%m-%d %H:%M:%S')   #This logging is always info, since for installation only
  for package in ['urllib3', 'requests', 'asyncio', 'PIL', 'matplotlib.font_manager']:
    logging.getLogger(package).setLevel(logging.WARNING)
  logging.info('Start 12 test: deterministic process')

  # start app and load project
  exampleData(True, None, 'research', '')
  backend = Backend('research')
  palette = Palette(None, 'dark_blue')
  comm = Communicate(backend, palette)
  window = Form(comm, {'_projectID': '', 'type': ['x0']})
  qtbot.addWidget(window)

  # change project to make it displayable
  projID = backend.output('x0').split('|')[-1].strip()
  doc   = backend.db.getDoc(projID)
  doc['comment'] = '# PASTA-ELN | The favorite ELN for experimental scientists\n - This is an example .eln output for sharing between different ELNs.'
  backend.editData(doc)
  backend.changeHierarchy(projID)
  dirName = backend.basePath/backend.cwd
  shutil.copy(Path(__file__).parent.parent/'pasta_eln'/'Resources'/'Icons'/'pasta512.png', dirName)
  backend.scanProject(DummyProgressBar(), projID)

  # move to top: directly in the database
  cmd = "UPDATE branches SET child='-1' WHERE path=='PastasExampleProject/pasta512.png'"
  backend.db.cursor.execute(cmd)
  backend.db.connection.commit()

  # window.change(projID,'')
    #change to pyside
    #form: create project
    #add folders (2) -> new item (copy from
    #
      # self.comm.backend.cwd = Path(self.comm.backend.db.getDoc(docID)['branch'][0]['path'])
      # docID = self.comm.backend.addData(docType, {'name':'new item'}, hierStack)['id']
      # self.comm.backend.cwd = Path(self.comm.backend.db.getDoc(docID)['branch'][0]['path'])
      # docID = self.comm.backend.addData(docType, {'name':'new item'}, hierStack)['id']

  verify(backend)
  return

def verify(backend): # Verify DB
  output = backend.checkDB(outputStyle='text')
  print(output)
  output = '\n'.join(output.split('\n')[8:])
  assert '**ERROR' not in output, 'Error in checkDB'
  assert len(output.split('\n')) == 5, 'Check db should have 5 more-less empty lines'
  return



""" Entire config dialog (dialog is blocking the main-window, as opposed to create a new widget-window)"""
from PySide6.QtWidgets import QDialog, QTabWidget, QVBoxLayout  # pylint: disable=no-name-in-module
from ..guiCommunicate import Communicate
from .configAuthors import ConfigurationAuthors
from .configGUI import ConfigurationGUI
from .configProjectGroup import ProjectGroup
from .configSetup import ConfigurationSetup


class Configuration(QDialog):
  """ Main class of entire config dialog """
  def __init__(self, comm:Communicate, startTab:str=''):
    """
    Initialization

    Args:
      backend (Pasta): backend, not communication
      startTab (str): tab to show initially
    """
    super().__init__()
    self.comm = comm
    self.setWindowTitle('PASTA-ELN configuration')
    self.setMinimumWidth(1200)

    # GUI elements
    mainL = QVBoxLayout(self)
    tabW = QTabWidget(self)
    mainL.addWidget(tabW)

    tabProjectGroup = ProjectGroup(self.comm, self.closeWidget)     # Project group. Restart app
    tabW.addTab(tabProjectGroup, 'Project group')

    tabGUI = ConfigurationGUI(self.comm, self.closeWidget)          # Misc configuration: e.g. theming. Restart app
    tabW.addTab(tabGUI, 'Appearance')

    tabAuthors = ConfigurationAuthors(self.comm, self.closeWidget)  # Author(s)
    tabW.addTab(tabAuthors, 'Author')

    tabSetup = ConfigurationSetup(self.comm, self.closeWidget)      # Setup / Troubleshoot Pasta
    tabW.addTab(tabSetup, 'Setup')

    # initialize when setup is called
    if startTab=='setup':
      tabW.setCurrentWidget(tabSetup)
      tabW.setTabEnabled(0, False)
      tabW.setTabEnabled(1, False)
      tabW.setTabEnabled(2, False)


  def closeWidget(self, restart:bool=True) -> None:
    """
    callback function to close widget, without restarting the entire app

    Args:
      restart (bool): if true, initialize
    """
    self.close()
    if restart:
      self.comm.backend.initialize()  #restart backend
    return



""" Utility function used by the dataverse module """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: utility_functions.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
from typing import Callable
from PySide6.QtCore import QDate, QSize
from PySide6.QtWidgets import (QBoxLayout, QComboBox, QDateTimeEdit, QFrame, QHBoxLayout, QLineEdit, QPushButton,
                               QSizePolicy, QVBoxLayout)
from pasta_eln.dataverse.utils import adjust_type_name


def create_push_button(button_display_text: str,
                       object_name: str,
                       tooltip: str,
                       parent_frame: QFrame,
                       parent_layout: QBoxLayout,
                       clicked_callback: Callable[[QBoxLayout], None]) -> QPushButton:
  """
  Creates a push button widget with specified properties.

  Args:
    button_display_text (str): The text to display on the button.
    object_name (str): The object name for the button.
    tooltip (str): The tooltip text for the button.
    parent_frame (QFrame): The parent QFrame object.
    parent_layout (QBoxLayout): The parent QBoxLayout object.
    clicked_callback (Callable[[QBoxLayout], None]): The callback function to execute when the button is clicked.


  Returns:
    QPushButton: The created push button widget.
  """
  push_button = QPushButton(parent=parent_frame)
  push_button.setText(button_display_text)
  push_button.setObjectName(object_name)
  push_button.setToolTip(tooltip)
  size_policy = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
  size_policy.setHorizontalStretch(0)
  size_policy.setVerticalStretch(0)
  size_policy.setHeightForWidth(push_button.sizePolicy().hasHeightForWidth())
  push_button.setSizePolicy(size_policy)
  push_button.setMinimumSize(QSize(100, 0))
  push_button.clicked.connect(lambda: clicked_callback(parent_layout))
  return push_button


def add_delete_button(parent_frame: QFrame,
                      parent: QBoxLayout,

                      enabled: bool = True) -> None:
  """
  Adds a delete button to the layout.

  Args:
      parent_frame (QFrame): The parent QFrame object.
      parent (QBoxLayout): The parent QBoxLayout object.
      enabled (bool, optional): Whether the button should be enabled or disabled.
      Defaults to True.

  Explanation:
      This function creates a QPushButton widget with the specified properties and adds it to the parent layout.
      The button is labeled as "Delete" and can be enabled or disabled based on the 'enabled' parameter.
      It is connected to a lambda function that calls the 'delete' function with the parent QBoxLayout as an argument.
  """
  delete_push_button = create_push_button(
    'Delete',
    'deletePushButton',
    'Delete this particular entry.',
    parent_frame,
    parent,
    delete_layout_and_contents)
  delete_push_button.setEnabled(enabled)
  parent.addWidget(delete_push_button)


def add_clear_button(parent_frame: QFrame,
                     parent: QBoxLayout) -> None:
  """
  Adds a clear button to the layout.

  Args:
      parent_frame (QFrame): The parent QFrame object.
      parent (QBoxLayout): The parent QBoxLayout object.

  Explanation:
      This function creates a QPushButton widget with the specified properties and adds it to the parent layout.
      The button is labeled as "Clear" and can be connected to a lambda function that calls the 'clear' function
      with the parent QBoxLayout as an argument.
  """
  button = create_push_button(
    'Clear',
    'clearPushButton',
    'Clear this particular entry.',
    parent_frame,
    parent,
    clear_layout_widgets,
  )
  parent.addWidget(button)


def create_date_time_widget(type_name: str,
                            type_value: str,
                            parent_frame: QFrame,
                            template_value: str | None = None) -> QDateTimeEdit:
  """
  Creates a date-time edit widget.

  Args:
      parent_frame (QFrame): The parent QFrame object.
      type_name (str): The name of the type.
      type_value (str): The initial value for the date-time edit.
      template_value (str | None, optional): The template value for the date-time edit. Defaults to None.

  Returns:
      QDateTimeEdit: The created date-time edit widget.

  Explanation:
      This function creates a QDateTimeEdit widget with the specified properties.
      The widget is set with a tooltip based on the type name and template value.
      It is assigned an object name based on the type name.
      The initial value is set for the date-time edit, and the display format is set to 'yyyy-MM-dd'.

  """
  date_time_edit = QDateTimeEdit(parent=parent_frame)
  date_time_edit.setSpecialValueText('No Value')
  date_time_edit.setMinimumDate(QDate.fromString('0100-01-01', 'yyyy-MM-dd'))
  adjusted_name = adjust_type_name(type_name)
  date_time_edit.setToolTip(
    f"Enter the {adjusted_name} value here. e.g. {template_value}, Minimum possible date is 0100-01-02")
  date_time_edit.setObjectName(f"{type_name}DateTimeEdit")
  date_time_edit.setDate(QDate.fromString(type_value, 'yyyy-MM-dd')
                         if type_value and type_value != 'No Value'
                         else QDate.fromString('0001-01-01', 'yyyy-MM-dd'))
  date_time_edit.setDisplayFormat('yyyy-MM-dd')
  return date_time_edit


def create_line_edit(type_name: str,
                     type_value: str,
                     parent_frame: QFrame,
                     template_value: str | None = None) -> QLineEdit:
  """
  Creates a line edit widget.

  Args:
      parent_frame (QFrame): The parent QFrame object.
      type_name (str): The name of the type.
      type_value (str): The initial value for the line edit.
      template_value (str | None, optional): The template value for the line edit. Defaults to None.

  Returns:
      QLineEdit: The created line edit widget.

  Explanation:
      This function creates a QLineEdit widget with the specified properties.
      The line edit is set with a placeholder text and a tooltip based on the type name and template value.
      It has a clear button enabled and is assigned an object name based on the type name.
      The initial value is set for the line edit.

  """
  entry_line_edit = QLineEdit(parent=parent_frame)
  adjusted_name = adjust_type_name(type_name)
  entry_line_edit.setPlaceholderText(f"Enter the {adjusted_name} here.")
  entry_line_edit.setToolTip(f"Enter the {adjusted_name} value here. e.g. {template_value}")
  entry_line_edit.setClearButtonEnabled(True)
  entry_line_edit.setObjectName(f"{type_name}LineEdit")
  entry_line_edit.setText(type_value or '')
  return entry_line_edit


def get_primitive_line_edit_text_value(primitive_vertical_layout: QVBoxLayout, widget_pos: int) -> str:
  """
  Retrieves the text value of a QLineEdit widget in a primitive vertical layout.

  Args:
      primitive_vertical_layout (QVBoxLayout): The vertical layout containing the QLineEdit widget.
      widget_pos (int): The position of the widget in the layout.

  Returns:
      str: The text value of the QLineEdit widget if found, otherwise None.

  Explanation:
      This function retrieves the text value of a QLineEdit widget located at the specified position in a primitive vertical layout.
      It first accesses the horizontal layout at the given position and then checks if the widget is a QLineEdit before returning its text value.
  """
  primitive_horizontal_layout = primitive_vertical_layout.itemAt(widget_pos).layout()
  if isinstance(primitive_horizontal_layout, QHBoxLayout):
    line_edit = primitive_horizontal_layout.itemAt(0).widget()
    if isinstance(line_edit, QLineEdit):
      return line_edit.text()
  return ''


def clear_layout_widgets(layout: QBoxLayout) -> None:
  """
  Clears the widgets in the given layout by resetting QLineEdit, QComboBox, and QDateTimeEdit widgets.

  Args:
    layout (QBoxLayout): The layout containing the widgets to be cleared.
  """
  if layout is None:
    return
  for widget_pos in reversed(range(layout.count())):
    if item := layout.itemAt(widget_pos):
      widget = item.widget()
      if isinstance(widget, QLineEdit):
        widget.clear()
      elif isinstance(widget, QComboBox):
        widget.setCurrentText('No Value')
      elif isinstance(widget, QDateTimeEdit):
        widget.setDate(QDate.fromString('01/01/0001', 'dd/MM/yyyy'))


def delete_layout_and_contents(layout: QBoxLayout) -> None:
  """
  Deletes the layout and its contents.

  Args:
      layout (QBoxLayout): The layout to delete.

  Explanation:
      This function deletes the given layout and its contents.
      It iterates over the widgets in the layout and removes them from their parent.
      Finally, it sets the layout's parent to None.

  """
  if layout is None:
    return
  for widget_pos in reversed(range(layout.count())):
    if item := layout.itemAt(widget_pos):
      item.widget().setParent(None)
  layout.setParent(None)



""" Provides an interface to interact with a database. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: base_database_api.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
import logging
from collections.abc import Callable
from typing import Union
from sqlalchemy import create_engine, or_, select
from sqlalchemy.orm import Session
from pasta_eln.database.error import Error
from pasta_eln.database.incorrect_parameter_error import IncorrectParameterError
from pasta_eln.database.models.base_model import BaseModel
from pasta_eln.database.models.config_model import ConfigModel
from pasta_eln.database.models.config_orm_model import ConfigOrmModel
from pasta_eln.database.models.data_hierarchy_model import DataHierarchyModel
from pasta_eln.database.models.data_hierarchy_orm_model import DataHierarchyOrmModel
from pasta_eln.database.models.main_orm_model import MainOrmModel
from pasta_eln.database.models.orm_model_base import OrmModelBase
from pasta_eln.database.models.project_model import ProjectModel
from pasta_eln.database.models.properties_orm_model import PropertiesOrmModel
from pasta_eln.database.models.upload_model import UploadModel
from pasta_eln.database.models.upload_orm_model import UploadOrmModel
from pasta_eln.database.orm_model_adapter import OrmModelAdapter
from pasta_eln.dataverse.database_names import DatabaseNames
from pasta_eln.dataverse.utils import generate_project_join_statement, log_and_create_error


class BaseDatabaseApi:
  """Handles database operations for various model types.

  This class provides methods to interact with the database, including creating,
  updating, retrieving, and counting models. It supports multiple model types and
  manages the conversion between base models and ORM models.

  Args:
      dataverse_db_path (str): The path to the Dataverse database.
      pasta_project_group_db_path (str): The path to the Pasta Project Group database.
  """

  def __init__(self,
               dataverse_db_path: str,
               pasta_project_group_db_path: str) -> None:
    self.logger = logging.getLogger('sqlalchemy.engine')
    self.logger.setLevel(logging.INFO)
    self.db_url_map: dict[DatabaseNames, str] = {}
    base_model_type = type[BaseModel | UploadModel | ConfigModel | ProjectModel | DataHierarchyModel]
    orm_model_type = type[
      OrmModelBase | UploadOrmModel | ConfigOrmModel | DataHierarchyOrmModel]
    self.model_mapping: dict[base_model_type, orm_model_type] = {
      UploadModel: UploadOrmModel,
      ConfigModel: ConfigOrmModel,
      DataHierarchyModel: DataHierarchyOrmModel
    }
    self.to_orm_converter_map: dict[base_model_type,  # type: ignore[valid-type, misc]
    Callable[base_model_type, orm_model_type]] = {
      UploadModel: OrmModelAdapter.get_orm_upload_model,
      ConfigModel: OrmModelAdapter.get_orm_config_model,
      ProjectModel: OrmModelAdapter.get_orm_project_model,
      DataHierarchyModel: OrmModelAdapter.get_orm_data_hierarchy_model
    }

    self.to_base_model_converter_map: dict[orm_model_type,  # type: ignore[valid-type, misc]
    Callable[orm_model_type, base_model_type]] = {
      UploadOrmModel: OrmModelAdapter.get_upload_model,
      ConfigOrmModel: OrmModelAdapter.get_config_model,
      DataHierarchyOrmModel: OrmModelAdapter.get_data_hierarchy_model
    }

    if isinstance(dataverse_db_path, str):
      self.db_url_map[DatabaseNames.DataverseDatabase] = f"sqlite:///{dataverse_db_path}"
    else:
      raise IncorrectParameterError(f"Database path must be a string: {dataverse_db_path}")
    if isinstance(pasta_project_group_db_path, str):
      self.db_url_map[DatabaseNames.PastaProjectGroupDatabase] = f"sqlite:///{pasta_project_group_db_path}"
    else:
      raise IncorrectParameterError(f"Database path must be a string: {pasta_project_group_db_path}")
    self.model_db_url_map: dict[type[UploadModel | ConfigModel | ProjectModel | DataHierarchyModel], str] = {
      UploadModel: self.db_url_map[DatabaseNames.DataverseDatabase],
      ConfigModel: self.db_url_map[DatabaseNames.DataverseDatabase],
      DataHierarchyModel: self.db_url_map[DatabaseNames.PastaProjectGroupDatabase],
      ProjectModel: self.db_url_map[DatabaseNames.PastaProjectGroupDatabase],
    }

  def create_and_init_database(self) -> None:
    """Creates and initializes the database.

    This function sets up the database by creating the necessary tables for the
    application. It logs the database creation process and ensures that the tables
    are created only if they do not already exist.

    Args:
        self: The instance of the class.

    Raises:
        SQLAlchemyError: If there is an error during the database creation process.
    """
    self.logger.info('Creating database at the location : %s',
                     self.db_url_map[DatabaseNames.DataverseDatabase])
    engine = create_engine(self.db_url_map[DatabaseNames.DataverseDatabase])
    (OrmModelBase.metadata.tables[ConfigOrmModel.__tablename__]
     .create(bind=engine, checkfirst=True))
    (OrmModelBase.metadata.tables[UploadOrmModel.__tablename__]
     .create(bind=engine, checkfirst=True))
    engine = create_engine(self.db_url_map[DatabaseNames.PastaProjectGroupDatabase])
    (OrmModelBase.metadata.tables[MainOrmModel.__tablename__]
     .create(bind=engine, checkfirst=True))
    (OrmModelBase.metadata.tables[PropertiesOrmModel.__tablename__]
     .create(bind=engine, checkfirst=True))
    (OrmModelBase.metadata.tables[DataHierarchyOrmModel.__tablename__]
     .create(bind=engine, checkfirst=True))

  def insert_model(self, data_model: Union[UploadModel, ConfigModel]) -> Union[UploadModel, ConfigModel]:
    """Inserts a data model into the database.

    This function takes a data model, converts it to the corresponding ORM model,
    and inserts it into the database. It logs the operation and returns the inserted
    model in its base form.

    Args:
        data_model(Union[UploadModel, ConfigModel]) : The data model to be inserted, which can be either an
        UploadModel or a ConfigModel.

    Returns:
        The inserted data model in its base form.

    Raises:
        SQLAlchemyError: If there is an error during the database insertion process.
    """
    self.logger.info('Populating document with data: %s in database: %s',
                     data_model,
                     self.db_url_map[DatabaseNames.DataverseDatabase])
    engine = create_engine(self.db_url_map[DatabaseNames.DataverseDatabase])
    model = self.to_orm_converter_map[type(data_model)](data_model)
    with Session(engine) as session:
      session.add(model)
      session.commit()
      session.flush()
      return self.to_base_model_converter_map[type(model)](model)

  def get_model(self, model_id: int | str,
                model_type: type[Union[UploadModel, ConfigModel, DataHierarchyModel, ProjectModel]]) -> Union[
    UploadModel, ProjectModel, ConfigModel, DataHierarchyModel, None]:
    """Retrieves a data model from the database based on its ID and type.

    This function fetches a specified model from the database using its ID and type.
    It logs the retrieval process and returns the model in its base form, or None if
    the model does not exist.

    Args:
        model_id (int | str): The ID of the model to retrieve, which can be an integer or a string.
        model_type (Type[Union[UploadModel, ConfigModel, DataHierarchyModel, ProjectModel]]):
            The type of the model to retrieve, which can be one of UploadModel, ConfigModel,
            DataHierarchyModel, or ProjectModel.

    Returns:
        Union[UploadModel, ProjectModel, ConfigModel, DataHierarchyModel, None]:
            The retrieved model in its base form, or None if the model does not exist.

    Raises:
        DatabaseError: If the model ID is empty or if the model type is not found.
    """
    self.logger.info('Retrieving data model with id: %s, type: %s', model_id, model_type)
    if not model_id:
      raise log_and_create_error(self.logger, Error, 'Model ID cannot be empty!')

    match model_type():
      case UploadModel() | ConfigModel() | DataHierarchyModel():
        engine = create_engine(self.model_db_url_map[model_type])
        with Session(engine) as session:
          db_model = session.get(self.model_mapping[model_type], model_id)
          return self.to_base_model_converter_map[self.model_mapping[model_type]](db_model) if db_model else None
      case ProjectModel():
        return self.get_project_model(model_id) if isinstance(model_id, str) else None
      case _:
        raise log_and_create_error(self.logger, Error, 'Model type not found!')

  def get_models(self, model_type: type[UploadModel | ConfigModel | DataHierarchyModel]) -> list[
    Union[UploadModel, ConfigModel, DataHierarchyModel]]:
    """Retrieves a list of models from the database based on the specified type.

    This function queries the database for models of the given type and returns them
    in a list. It logs the retrieval process and raises an error if the model type is
    not provided.

    Args:
        model_type (Type[UploadModel | ConfigModel | DataHierarchyModel]):
            The type of models to retrieve, which can be UploadModel, ConfigModel,
            or DataHierarchyModel.

    Returns:
        list[Union[UploadModel, ConfigModel, DataHierarchyModel]]:
            A list of models of the specified type retrieved from the database.

    Raises:
        DatabaseError: If the model type is not provided or if there is an error
        during the database retrieval process.
    """
    self.logger.info('Retrieving models from database, type: %s', model_type)
    if not model_type:
      raise log_and_create_error(self.logger, Error, 'Model Type cannot be empty!')
    stmt = select(self.model_mapping[model_type])
    engine = create_engine(self.model_db_url_map[model_type])
    with Session(engine) as session:
      models = session.scalars(stmt).all()
      return [self.to_base_model_converter_map[type(model)](model) for model in models]

  def get_project_models(self) -> list[ProjectModel]:
    """Retrieves a list of project models from the database.

    This function queries the database for all project models and returns them
    as a list. It logs the retrieval process and utilizes a session to execute
    the query and convert the results into project model instances.

    Returns:
        list[ProjectModel]: A list of project models retrieved from the database.

    Raises:
        SQLAlchemyError: If there is an error during the database retrieval process.
    """
    self.logger.info('Retrieving projects from database: %s', DatabaseNames.PastaProjectGroupDatabase)
    engine = create_engine(self.db_url_map[DatabaseNames.PastaProjectGroupDatabase])
    statement = generate_project_join_statement(None)
    with Session(engine) as session:
      return [OrmModelAdapter.get_project_model(r.tuple()) for r in session.execute(statement).fetchall()]

  def get_project_model(self, model_id: str) -> ProjectModel | None:
    """Retrieves a project model from the database using its ID.

    This function queries the database for a project model associated with the given
    model ID. It logs the retrieval process and raises an error if the model ID is
    not provided.

    Args:
        model_id (str): The ID of the project model to retrieve.

    Returns:
        ProjectModel | None: The project model retrieved from the database.

    Raises:
        DatabaseError: If the model ID is empty or if there is an error during
        the database retrieval process.
    """
    self.logger.info('Retrieving project from database: %s, model id: %s',
                     DatabaseNames.PastaProjectGroupDatabase,
                     model_id)
    if not model_id:
      raise log_and_create_error(self.logger, Error, 'Model ID cannot be empty!')
    engine = create_engine(self.db_url_map[DatabaseNames.PastaProjectGroupDatabase])
    statement = generate_project_join_statement(model_id)
    with Session(engine) as session:
      if first := session.execute(statement).fetchone():
        return OrmModelAdapter.get_project_model(first.tuple())
      return None

  def update_model(self, data_model: Union[UploadModel, ConfigModel]) -> None:
    """Updates an existing data model in the database.

    This function takes a data model, verifies its ID, and updates the corresponding
    record in the database. It logs the update process and raises an error if the
    model ID is not provided or if the model does not exist in the database.

    Args:
        data_model (Union[UploadModel, ConfigModel]): The data model to be updated,
        which can be either an UploadModel or a ConfigModel.

    Raises:
        DatabaseError: If the model ID is empty or if the model does not exist in
        the database during the update process.
    """
    model_type = type(data_model)
    self.logger.info('Updating data model with id: %s, type: %s', data_model.id,
                     model_type)
    if not data_model.id:
      raise log_and_create_error(self.logger, Error, 'Model ID cannot be empty!')
    engine = create_engine(self.model_db_url_map[model_type])
    with Session(engine) as session:
      db_model = self.to_orm_converter_map[model_type](data_model)
      if not (session.get(type(db_model),
                          db_model.id)):
        raise log_and_create_error(self.logger, Error, 'Model does not exist in database!')
      session.merge(db_model)
      session.commit()

  def get_paginated_models(self,
                           model_type: type[Union[UploadModel, ConfigModel, DataHierarchyModel]],
                           filter_term: str | None = None,
                           filter_fields: list[str] | None = None,
                           order_by_column: str | None = None,
                           page_number: int = 1,
                           limit: int = 10) -> list[type[UploadModel | ConfigModel | DataHierarchyModel]]:
    """Retrieves a paginated list of models from the database.

    This function fetches a specified number of models of a given type from the
    database, applying optional filters and sorting. It supports pagination to
    manage large datasets effectively.

    Args:
        model_type (Type[Union[UploadModel, ConfigModel, DataHierarchyModel]]):
            The type of models to retrieve, which can be UploadModel, ConfigModel,
            or DataHierarchyModel.
        filter_term (str | None): An optional term to filter the results.
        filter_fields (list[str] | None): An optional list of fields to apply the filter on.
        order_by_column (str | None): An optional column name to sort the results by.
        page_number (int): The page number to retrieve (default is 1).
        limit (int): The maximum number of results to return per page (default is 10).

    Returns:
        list[Type[UploadModel | ConfigModel | DataHierarchyModel]]:
            A list of models of the specified type retrieved from the database.

    Raises:
        DatabaseError: If the page number or limit is less than 1.
    """
    if page_number < 1:
      raise log_and_create_error(self.logger, Error, 'Page number cannot be less than 1!')
    if limit < 1:
      raise log_and_create_error(self.logger, Error, 'Limit cannot be less than 1!')
    engine = create_engine(self.model_db_url_map[model_type])
    query = select(self.model_mapping[model_type]).limit(limit).offset((page_number - 1) * limit)
    with Session(engine) as session:
      if filter_fields is None:
        filter_fields = self.model_mapping[model_type].get_table_columns()
      if filter_term:
        query = query.filter(
          or_(*[getattr(self.model_mapping[model_type], field).like(f"%{filter_term}%") for field in filter_fields]))
      if order_by_column:
        query = query.order_by(getattr(self.model_mapping[model_type], order_by_column).desc())
      models = session.execute(query).scalars().all()
      return [self.to_base_model_converter_map[type(model)](model) for model in models]

  def get_models_count(self,
                       model_type: type[Union[UploadModel, ConfigModel, DataHierarchyModel]]
                       ) -> int:
    """Retrieves the count of models of a specified type from the database.

    This function queries the database to count the number of records for the
    specified model type. It returns the total count, allowing for quick
    assessments of the number of entries in the database.

    Args:
        model_type (Type[Union[UploadModel, ConfigModel, DataHierarchyModel]]):
            The type of models for which to retrieve the count, which can be
            UploadModel, ConfigModel, or DataHierarchyModel.

    Returns:
        int: The count of models of the specified type in the database.

    Raises:
        KeyError: If the model type is not found in the model mapping.
    """
    engine = create_engine(self.model_db_url_map[model_type])
    with Session(engine) as session:
      return session.query(self.model_mapping[model_type]).count()



""" API for accessing an elabFTW server. That's API is inconvenient, complicated, ..."""
import base64
import copy
import json
import mimetypes
from pathlib import Path
from typing import Any, TypedDict
import requests  # only requirement; could be replaced with urllib to eliminate requirements


class ElabFTWApi:
  """ API for accessing an elabFTW server. That's API is inconvenient, complicated, ..."""

  def __init__(self, url:str='invalid', apiKey:str='', verifySSL:bool=True):
    '''
    initiate an elab instance to allow for syncing

    Args:
      url (str): url
      apiKey (str): API key
      veifySSL (bool): verify SSL certificate
    '''
    if url=='invalid':
      url = input('Please enter the url: ').strip()
      url = url if url.startswith('htt') else f'https://{url}'
      url = url if url.endswith('/')     else f'{url}/'
      print('Log-in to elabFTW server.')
      print(f'Go to website: {url}ucp.php?tab=4, enter a name and as permission: read/write. Create the API key')
      apiKey = input('Copy-paste the api-key: ').strip()
    class Param(TypedDict):
      """ class for parameters """
      headers: dict[str, str]
      verify: bool
      timeout: int
    # test server
    self.url = ''  #initialize: indicator if initialization successful
    self.headers = {'Content-type': 'application/json', 'Authorization': apiKey, 'Accept': 'text/plain'}
    self.param:Param = {'headers':self.headers, 'verify':verifySSL, 'timeout':10}
    try:
      response = requests.get(f'{url}info', **self.param)
      if response.status_code == 200:
        elabVersion = int(json.loads(response.content.decode('utf-8')).get('elabftw_version','0.0.0').split('.')[0])
        if elabVersion<5:
          print('**ERROR old elab-ftw version')
        else:
          self.url   = url
      else:
        print('|',response.status_code,'|')
        print('**ERROR not an elab-ftw server')
    except requests.ConnectionError:
      try:
        response = requests.get('https://www.google.com', headers={'Content-type': 'application/json'}, timeout=60)
        print('**ERROR not an elab-ftw server or cannot connect to that server.')
      except requests.ConnectionError:
        print('**ERROR: cannot connect to google. You are not online')
    return


  def touchEntry(self, entryType:str, content:dict[str,Any]={}) -> int:
    """
    create entry of type: experiment, item/resource, without much content

    Args:
      entryType (str): entryType to create, e.g. experiments, items, items_types
      content (dict): content to create

    Returns:
      int: elabFTW id
    """
    response = requests.post(self.url+entryType, data=json.dumps(content), **self.param)
    if response.status_code == 201:
      return int(response.headers['Location'].split('/')[-1])
    if response.status_code == 400:
      print(f"**ERROR occurred in touch of url '{entryType}': {json.loads(response.content.decode('utf-8'))['description']}")
    return -1


  def createEntry(self, entryType:str, content:str='') -> bool:
    """
    create entry of type: experiment, item/resource

    Args:
      entryType (str): entryType to create, e.g. experiments, items, items_types
      content (str): content to create

    Returns:
      bool: success of operation
    """
    raise NotImplementedError('Not implemented and tested')
    # response = requests.post(self.url+entryType, **self.param)
    # print("**TODO", content)
    # if response.status_code == 201:
    #   return True
    # print(f"**ERROR occurred in create of url {entryType}: {response.json}")
    # return False


  def readEntry(self, entryType:str, identifier:int=-1) -> list[dict[str,Any]]:
    """
    read entry or all entries (use identifier=-1 in the latter case)
    - can also be used to read custom command: do not use identifier and give custom command as entryType

    Args:
      entryType (str): entryType to create, e.g. experiments, items, items_types, teams
      identifier (int): elabFTW's identifier; list all if none is given

    Returns:
      dict: content read
    """
    url = f'{self.url}{entryType}' if identifier==-1 else f'{self.url}{entryType}/{identifier}'
    response = requests.get(url, **self.param)
    if response.status_code == 200:
      res = json.loads(response.content.decode('utf-8'))
      return res if identifier == -1 else [res]
    print(f"**ERROR occurred in get of url {entryType} / {identifier}")
    return [{}]

  def updateEntry(self, entryType:str, identifier:int, content:dict[str,Any]={}) -> bool:
    """
    update entry: experiment, item/resource

    Args:
      entryType (str): entryType to create, e.g. experiments, items, items_types
      identifier (int): elabFTW's identifier
      content (dict): content to update

    Returns:
      bool: success of operation
    """
    tags = content.pop('tags',[])
    response = requests.patch(f'{self.url}{entryType}/{identifier}', data=json.dumps(content), **self.param)
    if response.status_code != 200:
      return False
    # separate tags handling
    response = requests.get(f'{self.url}{entryType}/{identifier}/tags', **self.param)
    for tag in tags:
      response = requests.post(f'{self.url}{entryType}/{identifier}/tags', data=json.dumps({'tag':tag}), **self.param)
    return response.status_code == 201 if tags else True


  def deleteEntry(self, entryType:str, identifier:int) -> bool:
    """
    delete entry of type: experiment, item/resource

    Args:
      entryType (str): entryType to create, e.g. experiments, items, items_types
      identifier (int): elabFTW's identifier

    Returns:
      bool: success of operation
    """
    response = requests.delete(f'{self.url}{entryType}/{identifier}', **self.param)
    if response.status_code == 204:
      return True
    print(f"**ERROR occurred in delete of url {entryType}")
    return False


  def purgeExperimentsItems(self, areYouSure:bool=False) -> None:
    """ Remove all experiments and items on server
    - not used in Pasta

    Args:
      areYouSure (bool): safety mechanism to prevent accidental operation
    """
    if not areYouSure:
      return
    for entryType in ['experiments','items']:
      response = requests.get(f'{self.url}{entryType}?archived=on', **self.param)
      for identifier in [i['id'] for i in json.loads(response.content.decode('utf-8'))]:
        response = requests.delete(f'{self.url}{entryType}/{identifier}', **self.param)
        if response.status_code != 204:
          print(f'**ERROR purge delete {entryType}: {identifier}')
    return


  ### ---------------------------------------------
  ### LINKS
  ### ---------------------------------------------
  def createLink(self, entryType:str, identifier:int, targetType:str, linkTarget:int) -> bool:
    """
    create a link

    Args:
      entryType (str): entry type to modify (items, experiments)
      identifier (int): entry to change
      targetType (str): entry type to link to (items, experiments)
      linkTarget (int): target of the link

    Returns:
      bool: success of operation
    """
    response = requests.post(f'{self.url}{entryType}/{identifier}/{targetType}_links/{linkTarget}', **self.param)
    if response.status_code == 201:
      return True
    print(f"**ERROR occurred in create of url f'{self.url}{entryType}/{identifier}/{targetType}_links/{linkTarget}': {response.json}")
    return False



  ### ---------------------------------------------
  ### UPLOADS
  ### ---------------------------------------------
  def upload(self, entryType:str, identifier:int, content:str='', fileName:str='', jsonContent:str='', comment:str='') -> int:
    """
    upload a file

    Args:
      entryType (str): entryType to which to attach the upload
      identifier (int): elab's identifier
      content (str): base64 content, if given, it is used
      fileName (str): if content is not given, this filename is used
      jsonContent (str): if given, this json is content is used to create a json file
      comment (str): optional comment

    Returns:
      int: id of upload; -1 on failure
    """
    data:dict[str,Any] = {}
    if content.startswith('<?xml'):
      data = {'comment':comment, 'file': ('thumbnail.svg', content.encode(), 'image/svg')}
    elif content.startswith('data:image/png'):
      data = {'comment':comment, 'file': ('thumbnail.png', base64.b64decode(content[22:]) , 'image/png')}
    elif content.startswith('data:image/jpg'):
      data = {'comment':comment, 'file': ('thumbnail.jpg', base64.b64decode(content[22:]) , 'image/jpg')}
    elif fileName:
      with open(fileName,'rb') as fIn:
        try:
          mime = mimetypes.types_map[Path(fileName).suffix]
        except Exception:
          mime = 'application/octet-stream'
        data = {'comment':comment, 'file': (Path(fileName).name, fIn.read(), mime)}
    elif jsonContent:
      data = {'comment':comment, 'file': ('do_not_change.json', jsonContent.encode(), 'text/json')}
    else:  #default for fast testing
      data = {'comment':comment, 'file': ('README.md', b'Read me!\n', 'text/markdown')}
    headers = copy.deepcopy(self.headers)
    del headers['Content-type'] #will automatically become 'multipart/form-data'
    response = requests.post(f'{self.url}{entryType}/{identifier}/uploads', headers=headers,
                             files=data, verify=self.param['verify'], timeout=60)
    if response.status_code == 201:
      return int(response.headers['Location'].split('/')[-1])
    print(f"**ERROR occurred in upload of url '{entryType}/{identifier}': {json.loads(response.content.decode('utf-8'))['description']}")
    return -1


  def upLoadUpdate(self, entryType:str, identifier:int, uploadID:int, content:dict[str,Any]={}) -> bool:
    """
    update an upload

    Args:
      entryType (str): entry to create, e.g. experiments, items, items_types
      identifier (int): elabFTW's identifier
      uploadID (int): identifier of the upload
      content (dict): content to update

    Returns:
      bool: success of operation
    """
    url = f'{self.url}{entryType}/{identifier}/uploads/{uploadID}'
    print(url)
    response = requests.patch(url, data=json.dumps(content), **self.param)
    return response.status_code == 200


  def uploadDelete(self, entryType:str, identifier:int, uploadID:int) -> bool:
    """
    delete an upload

    Args:
      entryType (str): entryType to create, e.g. experiments, items, items_types
      identifier (int): elabFTW's identifier
      uploadID (int): identifier of the upload

    Returns:
      bool: success of operation
    """
    response = requests.delete(f'{self.url}{entryType}/{identifier}/uploads/{uploadID}', **self.param)
    if response.status_code == 204:
      return True
    print(f"**ERROR occurred in upload delete of url {entryType}/{identifier}/uploads/{uploadID}")
    return False


  def download(self, entryType:str, identifier:int, elabData:dict[str,str]) -> dict[str,Any]:
    """ Download a file, aka previous upload

    Args:
      entryType (str): entryType to create, e.g. experiments, items, items_types
      identifier (int): elabFTW's identifier
      elabData (dict): elabFTW's data of the upload

    Returns:
      str: downloaded content str or byte-array
    """
    url = f"{self.url}{entryType}/{identifier}/uploads/{elabData['id']}?format='binary'"
    response = requests.get(url, **self.param)
    if response.status_code == 200:
      if elabData['real_name']== 'do_not_change.json':
        return json.loads(response.content.decode('utf-8'))
      return {'data':response.content}
    print(elabData,response.status_code, url)
    return {}


  ### ---------------------------------------------
  ### USER / TEAM GROUPS
  ### ---------------------------------------------
  def readGroups(self, teamID:int, groupID:int=-1) -> list[dict[str,Any]]:
    """ List all groups or just one (not used in Pasta)

    Args:
      teamID (int): elabFTW's identifier of the team
      groupID (int): elabFTW's identifier of the group

    Returns:
      list: list of reply
    """
    url = f"{self.url}teams/{teamID}/teamgroups" if groupID==-1 else f"{self.url}teams/{teamID}/teamgroups/{groupID}"
    response = requests.get(url, **self.param)
    if response.status_code == 200:
      res = json.loads(response.content.decode('utf-8'))
      return res if groupID == -1 else [res]
    print(f"**ERROR occurred in get of url: {url}")
    return [{}]



""" Adds a new compound entry. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: primitive_compound_frame.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

import logging
from typing import Any
from PySide6.QtWidgets import QFrame
from pasta_eln.GUI.dataverse.data_type_class_base import DataTypeClass
from pasta_eln.GUI.dataverse.data_type_class_context import DataTypeClassContext
from pasta_eln.GUI.dataverse.data_type_class_factory import DataTypeClassFactory
from pasta_eln.GUI.dataverse.data_type_class_names import DataTypeClassName
from pasta_eln.GUI.dataverse.metadata_frame_base import MetadataFrame
from pasta_eln.GUI.dataverse.primitive_compound_controlled_frame_base import Ui_PrimitiveCompoundControlledFrameBase


class PrimitiveCompoundFrame(Ui_PrimitiveCompoundControlledFrameBase, MetadataFrame):
  """
  Adds a new compound entry.

  Explanation:
      This method adds a new compound entry to the UI.
      It creates a new layout for the entry and adds the appropriate UI elements based on the types dictionary.

  """

  def __new__(cls, *_: Any, **__: Any) -> Any:
    """
    Creates a new instance of the PrimitiveCompoundFrame class.

    Explanation:
        This method creates a new instance of the PrimitiveCompoundFrame class.

    Args:
        *_: Variable length argument list.
        **__: Arbitrary keyword arguments.

    Returns:
        Any: The new instance of the PrimitiveCompoundFrame class.
    """
    return super().__new__(cls)

  def __init__(self, meta_field: dict[str, Any]) -> None:
    """
    Initializes the PrimitiveCompoundFrame.

    Explanation:
        This method initializes the PrimitiveCompoundFrame class.
        It sets up the UI and initializes the types dictionary.

    Args:
        meta_field (dict[str, Any]): The dictionary containing the metadata field information.
    """
    self.logger: logging.Logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
    self.instance: QFrame = QFrame()
    super().setupUi(self.instance)
    super().__init__(self.instance)
    self.meta_field: dict[str, Any] = meta_field
    self.data_type_class_factory: DataTypeClassFactory = DataTypeClassFactory(
      DataTypeClassContext(self.mainVerticalLayout, self.addPushButton, self.instance, meta_field)
    )
    self.data_type: DataTypeClass = self.data_type_class_factory.make_data_type_class(
      DataTypeClassName(meta_field['typeClass'])
    )
    self.addPushButton.clicked.connect(self.add_button_click_handler)
    self.load_ui()

  def load_ui(self) -> None:
    """
    Loads the UI based on the meta_field information.

    Args:
        self: The instance of the class.

    Explanation:
        This method loads the UI based on the meta_field information.
        It handles different cases based on the 'typeClass' value in the meta_field dictionary.
        - For 'primitive' type, it disables the addPushButton if 'multiple' is False and populates the primitive entry.
        - For 'compound' type, it handles the cases of 'multiple' being True or False.
          - If 'multiple' is True and the 'value' is empty, it creates an empty entry and populates the compound entry.
          - If 'multiple' is True and the 'value' is not empty, it populates the compound entry for each compound in the 'value'.
          - If 'multiple' is False and the 'value' is not empty, it populates the compound entry with the 'value'.
          - If 'multiple' is False and the 'value' is empty, it creates an empty entry and populates the compound entry.
        - For any other 'typeClass' value, it logs an error for unknown typeClass.

    """
    self.logger.info('Loading UI for %s', self.meta_field)
    self.data_type.populate_entry()

  def add_button_click_handler(self) -> None:
    """
        Adds a new entry based on the meta_field information.

        Explanation:
            This method adds a new entry based on the meta_field information.
            It handles different cases based on the 'typeClass' value in the meta_field dictionary.
            - For 'primitive' type, it populates the primitive horizontal layout with the specified values.
            - For 'compound' type, it creates a new compound horizontal layout and populates it with the specified values.
            - For any other 'typeClass' value, it logs an error for unknown typeClass.

    """
    self.logger.info('Adding new entry of type %s, name: %s',
                     self.meta_field.get('typeClass', ''),
                     self.meta_field.get('typeName', ''))
    if not self.meta_field.get('multiple'):
      self.logger.error('Add operation not supported for non-multiple entries')
      return
    self.data_type.add_new_entry()

  def save_modifications(self) -> None:
    """
    Saves the changes made in UI elements to the meta_field.

    Args:
        self: The instance of the class.

    Explanation:
        This method saves the changes made to the meta_field.
        It handles different cases based on the 'typeClass' value in the meta_field dictionary.
        - For 'primitive' type, it updates the 'value' attribute of the meta_field.
        - For 'compound' type, it updates the 'value' attribute of the meta_field based on the 'valueTemplate'.
        - For any other 'typeClass' value, it logs an error for unsupported typeClass.
    """
    self.logger.info('Saving changes to meta_field for type, name: %s, class: %s',
                     self.meta_field.get('typeName'),
                     self.meta_field.get('typeClass'))
    self.data_type.save_modifications()



""" Represents the ORM model for the upload table in the database. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: upload_orm_model.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

from typing import Optional
from sqlalchemy.orm import Mapped, mapped_column
from pasta_eln.database.models.orm_model_base import OrmModelBase


class UploadOrmModel(OrmModelBase):
  """Represents the ORM model for the upload table in the database.

  This class defines the structure of the upload model used in the application,
  including fields for ID, data type, project name, project document ID, status,
  creation date, finished date, log, and dataverse URL. It provides a method to
  retrieve the names of the table columns for database operations.
  """
  __tablename__ = 'upload'
  id: Mapped[int] = mapped_column(primary_key=True)
  data_type: Mapped[Optional[str]]
  project_name: Mapped[Optional[str]]
  project_doc_id: Mapped[Optional[str]]
  status: Mapped[Optional[str]]
  created_date_time: Mapped[Optional[str]]
  finished_date_time: Mapped[Optional[str]]
  log: Mapped[Optional[str]]
  dataverse_url: Mapped[Optional[str]]

  @classmethod
  def get_table_columns(cls) -> list[str]:
    """Retrieves the list of column names for the upload table.

    This method returns a list of strings representing the names of the columns
    defined in the upload model. It is useful for database operations that
    require knowledge of the table structure.

    Returns:
        list[str]: A list of column names for the upload table.
    """
    return ['id', 'data_type', 'project_name', 'project_doc_id', 'status', 'created_date_time', 'finished_date_time',
            'log', 'dataverse_url']
